#Fonctions pour la sortie en lettres:
#Fonction EnToutesLettres par Guy Lapalme, légèrement modifiée par Francis pour accomoder le genre
import re,sys

def enToutesLettres(s,lang):
    en=lang=="en"
    trace=False  # utile pour la mise au point

    # expressions des unités pour les "grands" nombres >1000 
    # expressions donnent les formes [{singulier, pluriel}...]
    #  noms de unités selon l'échelle courte présentée dans le Guide Antidote
    # elle diffère de celle présentée dans http:#villemin.gerard.free.fr/TABLES/NbLettre.htm
    unitesM=[   {"sing":"mille"         ,"plur":"mille"}        # 10^3
               ,{"sing":"un million"    ,"plur":"millions"}     # 10^6
               ,{"sing":"un milliard"   ,"plur":"milliards"}    # 10^9
               ,{"sing":"un trillion"   ,"plur":"trillions"}    # 10^12
               ,{"sing":"un quatrillion","plur":"quatrillions"} # 10^15
               ,{"sing":"un quintillion","plur":"quintillions"} # 10^18
                ]
    unitsM =[   {"sing":"one thousand"      ,"plur":"thousand"}    # 10^3
               ,{"sing":"one million"       ,"plur":"million"}     # 10^6
               ,{"sing":"one billion"       ,"plur":"billion"}     # 10^9
               ,{"sing":"one trillion"      ,"plur":"trillion"}    # 10^12
               ,{"sing":"one quatrillion"   ,"plur":"quatrillion"} # 10^15
               ,{"sing":"one quintillion"   ,"plur":"quintillion"} # 10^18
                ]

    maxLong=21  # longueur d'une chaîne de chiffres traitable (fixé par la liste unitesM)

    # séparer une chaine en groupes de trois et complétant le premier groupe avec des 0 au début
    def splitS(s):
        if len(s)>3:
            return splitS(s[:-3])+[s[-3:]]
        elif len(s)==1: s="00"+s
        elif len(s)==2: s="0"+s
        return [s]

    # est-ce que tous les triplets d'une liste correspondent à 0 ?
    def tousZero(ns):
        if len(ns)==0:return True;
        return ns[0]=="000" and tousZero(ns[1:])

    # création d'une liste de triplets de chiffres
    def grouper(ns): # ns est une liste de chaines de 3 chiffres
        l=len(ns)
        if trace: print("grouper:"+str(l)+":"+ns)
        head=ns[0]
        if l==1:return centaines(head)
        tail=ns[1:]
        if head=="000": return grouper(tail)
        uM=unitsM if en else unitesM
        return (uM[l-2]["sing"] if head=="001" else grouper([head])+" "+uM[l-2]["plur"])\
                +" "+("" if tousZero(tail) else grouper(tail))

    # traiter un nombre entre 0 et 999
    def centaines(ns): # ns est une chaine d'au plus trois chiffres
        if trace:print("centaines:"+ns)
        if len(ns)==1: return unites(ns)
        if len(ns)==2: return dizaines(ns)
        c=ns[0]       # centaines
        du=ns[1:] # dizaines+unités
        if c=="0": return dizaines(du)
        cent="hundred" if en else "cent"
        if du=="00":
            if c=="1":return ("one " if en else "")+cent
            return unites(c)+" "+cent+("" if en else "s")
        if c=="1": return ("one " if en else "")+cent+" "+dizaines(du)
        return unites(c)+" "+cent+(" and " if en else" ")+dizaines(du)

    # traiter un nombre entre 10 et 99
    def dizaines(ns):# ns est une chaine de deux chiffres
        if trace: print("dizaines:",ns);
        d=ns[0] # dizaines
        u=ns[1] # unités
        if d=="0": return unites(u)
        elif d=="1":
            if en: 
                return ["ten","eleven","twelve","thirteen","fourteen","fifteen","sixteen","seventeen","eighteen","nineteen"][int(u)]
            else:
                return["dix","onze","douze","treize","quatorze","quinze","seize","dix-sept","dix-huit","dix-neuf"][int(u)]
        elif d in "23456":
            tens = (["twenty","thirty","forty","fifty","sixty"] if en else 
                    ["vingt","trente","quarante","cinquante","soixante"])[int(d)-2]
            if u=="0": return tens
            return tens + ( ("-one" if en else " et un") if u=="1" else ("-"+unites(u)))
        elif d=="7":
            if u=="0": return "seventy" if en else "soixante-dix"
            return ("seventy-"+unites(u))if en else ("soixante-"+dizaines("1"+u))
        elif d=="8":
            if u=="0": return "eighty" if en else "quatre-vingts"
            return ("eighty-" if en else "quatre-vingt-")+unites(u)
        elif d=="9":
            if u=="0": return "ninety" if en else "quatre-vingt-dix"
            return ("ninety-"+unites(u))if en else ("quatre-vingt-"+dizaines("1"+u))
        else:
            print("strange dizaine:",ns,file=sys.stderr)

    # traiter un chiffre entre 0 et 10
    def unites(u): # u est une chaine d'un chiffre
        if en: return ["zero","one","two","three","four","five","six","seven","eight","nine"][int(u)]
        return ["zéro","un","deux","trois","quatre","cinq","six","sept","huit","neuf"][int(u)]
    
#/ début de l'exécution de la fonction
    if isinstance(s,int):s=str(s) # convertir un nombre en chaîne
    if not re.match(r"^-?\d+$",s):
        print("nombreChaineEnLettres ne traite que des chiffres:"+s,file=sys.stderr)
    neg=False
    if s[0]=="-":
        neg=True
        s=s[1:]
    if len(s)>maxLong:
        print("nombreChaineEnLettres ne traite que les nombres d'au plus "+str(maxLong)+" chiffres:"+s,sys.stderr)
    res=grouper(splitS(s))
    if neg: res=("minus " if en else "moins ")+res
    return res.strip()


# si l'orthographe française rectifiée est demandée, appliquer cette fonction à la sortie
# de enToutesLettres() pour mettre des tirets à la place des espaces partout dans le nombre...
# def rectifiee(s){
#     return s.replace(/ /g,"-");
# }

# écriture des nombres ordinaux   #GL

# rules taken from https:#www.ego4u.com/en/cram-up/vocabulary/numbers/ordinal
ordEnExceptions={"one":"first","two":"second","three":"third","five":"fifth",
                 "eight":"eighth","nine":"ninth","twelve":"twelfth"}
# règles tirées de https:#francais.lingolia.com/fr/vocabulaire/nombres-date-et-heure/les-nombres-ordinaux
ordFrExceptions={"un":"premier","une":"première","cinq":"cinquième","neuf":"neuvième"}

def ordinal(s,lang,gender):
    en = lang=="en"
    s=enToutesLettres(s,lang)
    if s=="zéro" or s=="zero": return s
    m=re.match(r"(.*?)(\w+)$",s)
    lastWord=m[2]
    if en: 
        if lastWord in ordEnExceptions: return m[1]+ordEnExceptions[lastWord]
        if s[-1]=="y": return s[:-1]+"ieth"; # added from the reference
        return s+"th"
    else:
        if s == "un":return "première" if gender=="f" else "premier"
        if s.endswith("et un"): return s+"ième"
        if lastWord in ordFrExceptions: return m[1]+ordFrExceptions[lastWord]
        if s[-1]=="e" or s.endswith("quatre-vingts"): return s[:-1]+"ième";
        return s+"ième"

if __name__ == '__main__':
    print(enToutesLettres("1234567","en"))
    print(enToutesLettres("1234567","fr"))
    print(enToutesLettres("-45", "en"))
    print(enToutesLettres("300", "fr"))
    print(ordinal(34,"en","f"))
    for n in range(7,100000, 397):
        print(enToutesLettres(str(n), "en"))
        print(enToutesLettres(str(n), "fr"))
        




from .Terminal import  N,A,Pro,D,Adv,V,P,C,DT,NO,Q
from .Phrase import NP,AP,VP,AdvP,PP,CP,S,SP
from .Lexicon import loadEn, loadFr

# create a list of elements [a,b,c] => "a, b $conj c" 
def makeDisj(conj,elems):     
    args=[C(conj)]+[Q(e) for e in elems]
    return CP(*args)

warnings = {
    "bad parameter":
        {"en":lambda good,bad: # the parameter should be $good, not $bad
            S(NP(D("the"),N("parameter")),
              VP(V("be").t("ps"),Q(good).a(","),Adv("not"),Q(bad))).typ({"mod":"nece"}),
         "fr":lambda good,bad: # le paramètre devrait être $good, pas $bad
            S(NP(D("le"),N("paramètre")),
              VP(V("être").t("c"),Q(good).a(","),Adv("non"),Q(bad))).typ({"mod":"nece"})},
    "bad application":
        {"en":lambda info,goods,bad: # $info should be applied to $good, not to $bad
            S(Q(info),VP(V("apply").t("ps"),
                         PP(P("to"),makeDisj("or",goods)).a(","),Adv("not"),PP(P("to"),Q(bad))))
               .typ({"mod":"nece","pas":True}),
         "fr":lambda info,goods,bad: # $info devrait être appliqué à $good, non à $bad.
            S(Q(info),VP(V("appliquer").t("c"),
                         PP(P("à"),makeDisj("ou",goods)).a(','),Adv("non"),PP(P("à"),Q(bad))))
              .typ({"mod":"nece","pas":True})},
    "bad position":
        {"en":lambda bad,limit : # $bad should be smaller than $limit.
            S(Q(bad),VP(V("be").t("ps"),A("small").f("co"),P("than"),Q(limit))).typ({"mod":"nece"}),
         "fr":lambda bad,limit : # $bad devrait être plus petit que $limit.
            S(Q(bad),VP(V("être").t("c"),A("petit").f("co"),Pro("que"),Q(limit))).typ({"mod":"nece"})},
    "bad const for option":
        {"en":lambda option,constType,allowedConsts : 
            # option $option is applied to $constType, but it should be to $allowedConsts.
              CP(C("but"),
                 VP(V("apply"),NP(N("option"),Q(option)),PP(P("to"),Q(constType))).typ({"pas":True}).a(","),
                 SP(Pro("I"),VP(V("be").t("ps"),PP(P("to"),makeDisj("or",allowedConsts)))).typ({"mod":"nece"})
              ),
         "fr":lambda option,constType,allowedConsts :
              #  l'option $option est appliquée à $constType, mais elle devrait l'être à $allowedConsts
              CP(C("mais"),
                 VP(V("appliquer"),NP(D("le"),N("option"),Q(option)),PP(P("à"),Q(constType)))
                    .typ({"pas":True}).a(","),
                 SP(Pro("je").g("f"),
                    VP(Pro("elle").c("acc"),
                       V("être").t("c"),PP(P("à"),makeDisj("ou",allowedConsts)))).typ({"mod":"nece"})
              )},
    "ignored value for option":
        {"en":lambda option,bad : # $bad: bad value for option $option is ignored.
            S(Q(bad).a(":"),
              VP(V("ignore"),NP(D("this"),A("bad"),N("value"),
                                PP(P("for"),N("option"),Q(option)))).typ({"pas":True})),
         "fr":lambda option,bad :  # $bad : cette mauvaise valeur pour l'option $option est ignorée
            S(Q(bad).a(":"),
              VP(V("ignorer"),NP(D("ce"),A("mauvais"),N("valeur"),
                                 PP(P("pour"),D("le"),N("option"),Q(option)))).typ({"pas":True}))},
    "unknown type":
        {"en":lambda key,allowedTypes : # illegal type: $key, it should be $allowedTypes.
            S(NP(A("illegal"),N("type").a(":"),Q(key)).a(","),
              VP(Pro("I"),V("be").t("ps"),makeDisj("or",allowedTypes))).typ({"mod":"nece"}),
         "fr":lambda key,allowedTypes : # type illégal : $key, il devrait être $allowedTypes.
            S(NP(N("type"),A("illégal").a(":"),Q(key)).a(","),
              SP(Pro("je"),VP(V("être").t("c"),makeDisj("ou",allowedTypes))).typ({"mod":"nece"}))},
    "no value for option":
        {"en":lambda option,validVals : # no value for option $option should be one of $validVals.
            S(NP(Adv("no"),N("value"),PP(P("for"),N("option"),Q(option))),
              VP(V("be").t("ps"),Pro("one"),PP(P("of"),Q(validVals)))).typ({"mod":"nece"}),
         "fr":lambda option,validVals : # aucune valeur pour l'option $option, devrait être une parmi $validVals.
            S(NP(D("aucun"),N("valeur"),PP(P("pour"),D("le"),N("option"),Q(option))).a(","),
              VP(V("être").t("c"),D("un").g("f"),PP(P("parmi"),Q(validVals)))).typ({"mod":"nece"})},
    "not found":
        {"en":lambda missing,context : # no $missing found in $context.
            S(AdvP(Adv("no"),Q(missing)),VP(V("find").t("pp"),PP(P("in"),Q(context)))),
         "fr":lambda missing,context : # aucun $missing trouvé dans $context.
            S(D("aucun"),Q(missing),VP(V("trouver").t("pp"),PP(P("dans"),Q(context))))},
    "bad ordinal":
        {"en":lambda value : # cannot realize $value as ordinal.
            S(VP(V("realize"),Q(value),AdvP(Adv("as"),N("ordinal")))).typ({"neg":True,"mod":"poss"}),
         "fr":lambda value : # $value ne peut pas être réalisé comme un ordinal.
            S(Q(value),VP(V("réaliser"),AdvP(Adv("comme"),NP(D("un"),N("ordinal")))))
              .typ({"neg":True,"mod":"poss","pas":True})},
    "bad number in word":
        {"en":lambda value : # cannot realize $value in words.
            S(VP(V("realize"),Q(value),PP(P("in"),N("word").n("p")))).typ({"neg":True,"mod":"poss"}),
         "fr":lambda value :# $value ne peut pas être réalisé en mots.
            S(VP(Q(value),V("réaliser"),PP(P("en"),NP(N("mot").n("p"))))).typ({"neg":True,"mod":"poss","pas":True})},
    "no French contraction":
        {"en":lambda  : # contraction is ignored in French.
            S(VP(V("ignore"),NP(N("contraction")),PP(P("in"),N("French")))).typ({"pas":True}),
         "fr":lambda  : # la contraction est ignorée en français.
            S(VP(V("ignorer"),NP(D("le"),N("contraction")),PP(P("en"),N("français")))).typ({"pas":True})},
    "morphology error":
        {"en":lambda info : # morphology error: $info.
            S(NP(N("morphology"),N("error")).a(":"),Q(info)),
         "fr":lambda info : # erreur de morphologie : $info.
            S(NP(N("erreur"),PP(P("de"),N("morphologie"))).a(":"),Q(info))},
    "not implemented":
        {"en":lambda info : # $info is not implemented.
            S(Q(info),VP(V("implement"))).typ({"neg":True,"pas":True}),
         "fr":lambda info : # $info n'est pas implémenté.
            S(Q(info),VP(V("implémenter"))).typ({"neg":True,"pas":True})},
    "not in lexicon":
        {"en":lambda lang,altPos=None : # not found in lexicon.
            S(Adv("not"),V("find").t("pp"),PP(P("in"),A("English" if lang=="en" else "French"),N("lexicon")),
              Q("") if altPos is None else AdvP(Adv("but"),V("exist"),Adv("as"),makeDisj("or",altPos))),
         "fr":lambda lang,altPos=None : # absent du lexique.
            S(AP(A("absent"),PP(P("de"),NP(D("le"),N("lexique"),A("anglais" if lang=="en" else "français")))),
              Q("") if altPos is None else AdvP(Adv("mais"),V("exister"),Adv("comme"),makeDisj("ou",altPos)))},
    "no appropriate pronoun":
        {"en":lambda  :S(VP(V("find").t("ps"),NP(D("a"),A("appropriate"),N("pronoun")))
                         ).typ({"neg":True,"pas":True,"mod":"poss"}),
         "fr":lambda  :S(VP(V("trouver").t("pc"),NP(D("un"),A("adéquat"),N("pronom")))
                         ).typ({"neg":True,"pas":True,"mod":"poss"})
        },
    "both tonic and clitic":
        {"en":lambda  :# tn(..) and c(..) cannot be used together, tn(..) is ignored.
             S(CP(C("and"),Q("tn(..)"),Q("c(..)")),VP(V("use").n("p"),Adv("together"))
                  .typ({"neg":True,"pas":True,"mod":"poss"}).a(","),
               Q("tn(..)"),VP(V("ignore")).typ({"pas":True})),
         "fr":lambda  :# tn(..) et c(..) utilisés ensemble, tn(..) est ignoré.
             S(SP(CP(C("et"),Q("tn(..)"),Q("c(..)")),VP(V("utiliser").t("pp").n("p"),Adv("ensemble"))).a(","),
               SP(Q("tn(..)"),VP(V("ignorer")).typ({"pas":True})))
        },
    "bad Constituent":
        {"en":lambda rank,type : # the $rank parameter is not Constituent.
            S(NP(D("the"),Q(rank),N("parameter")),
              VP(V("be"),Q("Constituent"),Adv("but"),Q(type))).typ({"neg":True}),
         "fr":lambda rank,type : # le $rank paramètre n'est pas Constituent.
            S(NP(D("le"),Q(rank),N("paramètre")),
              VP(V("être"),Q("Constituent"),Adv("mais"),Q(type))).typ({"neg":True})},
    "bad Dependent":
        {"en":lambda rank,type :  #   the $rank parameter is not Dependent but $type.
            S(NP(D("the"),Q(rank),N("parameter")),
              VP(V("be"),Q("Dependent"),Adv("but"),Q(type))).typ({"neg":True}),
         "fr":lambda rank,type :   #   le $rank paramètre n'est pas Dependent mais $type.
            S(NP(D("le"),Q(rank),N("paramètre")),
              VP(V("être"),Q("Dependent"),Adv("mais"),Q(type))).typ({"neg":True})} ,
    "Dependent needs Terminal":
        {"en":lambda type:  #   the first parameter of Dependent is not Terminal but $type.
            S(NP(D("the"),A("first"),N("parameter"),PP(P("of"),Q("Dependent"))),
              VP(V("be"),Q("Terminal"),Adv("but"),Q(type))).typ({"neg":True}),
         "fr":lambda type:  #   le premier paramètre du Dependent n'est pas Terminal mais $type.
            S(NP(D("le"),A("premier"),N("paramètre"),PP(P("de"),NP(D("le"),Q("Dependent")))),
              VP(V("être"),Q("Terminal"),Adv("mais"),Q(type))).typ({"neg":True})},
    "bad number of parameters":
        {"en":lambda termType,number : # $termType accepts one parameter, but has $number.
             S(Q(termType),VP(V("accept"),NP(D("a"),A("single"),N("parameter"))).a(","),
               SP(C("but"),Pro("I"),VP(VP(V("have"),Q(number))))),
         "fr":lambda termType,number : # $termType accepte un seul paramètre, mais en a $number.
             S(Q(termType),VP(V("accepter"),NP(D("un"),A("seul").pos("pre"),N("paramètre"))).a(","),
               SP(C("mais"),Pro("je"),VP(VP(Pro("en"),V("avoir"),Q(number)))))},
    "Dependent without params":
        {"en":lambda: # Dependent without parameter
             S(Q("Dependent"),PP(P("without"),N("parameter"))),
         "fr":lambda: # Dependent sans paramètre.
             S(Q("Dependent"),PP(P("sans"),N("paramètre")))},
    "bad lexicon table":
        {"en":lambda lemma,ending : # error in lexicon table number: $lemma should end with $ending
            S(NP(N("error"),P("in"),N("lexicon"),N("table"),N("number")).a(":"),
              SP(Q(lemma),VP(V("end"),PP(P("with"),Q(ending)))).typ({"neg":True})),
         "fr":lambda lemma,ending : # erreur de numéro de table dans le lexique: $lemma devrait terminer par $ending
            S(NP(N("erreur"),P("de"),N("numéro"),P("de"),N("table"),P("dans"),NP(D("le"),N("lexique"))).a(":"),
              SP(Q(lemma),VP(V("terminer"),PP(P("par"),Q(ending)))).typ({"neg":True}))},
    "bad language":{
        "en":lambda lang: # language should be "en" or "fr", it will be $lang
            S(N("language"),VP(V("be").t("ps"),CP(C("or"),Q("en").en("'"),Q("fr").en("'"))).typ({"mod":"nece"}).a(","),
              Pro("it"),VP(V("be").t("f"),Q(lang))),
        "fr":lambda lang: # langage devrait être "en" ou "fr", ce sera $lang
            S(N("langage"),VP(V("être").t("c"),CP(C("ou"),Q("en").en('"'),Q("fr").en('"'))).typ({"mod":"nece"}).a(","),
              Pro("ce"),VP(V("être").t("f"),Q(lang))),
        },
    "ignored reflexive":
        {"en":lambda pat: # cannot be reflexive, only $pat
            S(VP(V("be"),A("reflexive")).typ({"mod":"poss","neg":True}),Adv("only"),makeDisj("or",pat)),
         "fr":lambda pat: # ne peut être réflexif, seulement $pat
            S(VP(V("être"),A("réflexif")).typ({"mod":"poss","neg":True}),
              AdvP(Adv("seulement"),makeDisj("ou",pat)) if len(pat)>0 else None),
         },
    "inconsistent dependents within a coord":
        {"en":lambda  expected,found: #  $expected expected within this coord, but $found was found
            S(Q(expected),VP(V("expect").t("pp"),PP(P("within"),NP(D("this"),Q("coord")))),
              SP(C("but"),Q(found),V("be").t("ps"),V("find").t("pp"))),
         "fr": lambda expected,found: # toutes les dépendances d'un coord devraient être $expected, mais $found a été rencontré
            S(NP(A("tout"),D("le"),N("dépendance").n("p"),P("de"),D("un"),Q("coord")),
              VP(V("devoir").t("cp"),V("être").t("b"),Q(expected)),
              SP(C("mais"),Q(found),V("être").t("pc"),V("rencontrer").t("pp")))},
    "user-warning": # user specific message, either a String or a Constituent that will be realized
        {"en":lambda mess: Q(mess),
         "fr":lambda mess: Q(mess) },
}

def warning(self,args):
    if self.isEn():
        lang="en"
        loadEn()
    else:
        lang="fr"
        loadFr()
    args=list(args)
    mess=args.pop(0)
    if mess not in warnings:
        self.error("warn called with an unknown error message:"+str(mess))        
    messS=self.me()+":: "+ warnings[mess][lang](*args).cap(False).realize() # realize the warning
    return messS

def test_warnings():
    args=["A","B","C","D","E","F"]
    for w in warnings.keys():
        print(w)
        # determine the number of argument for the function
        nbArgs=warnings[w]["en"].__code__.co_argcount
        callArgs=[w]+args[:nbArgs] if w!="user-warning" else [w,Q("warning defined by the user")]
        loadEn()
        print(warning(NP(D("a"),N("error")),callArgs))
        loadFr()
        print(warning(NP(D("un"),N("erreur")),callArgs))
        print("---")
   


## Python version of jsRealB
##   the organization parallels the one of JavaScript version (version 4.0)
##   in fact, the JavaScript version was sometimes revised in order to make it
##   similar
## Guy Lapalme, December 2021

# select a random element in a list useful to have some variety in the generated text
#  if the first argument is a list, selection is done within the list
#  otherwise the selection is among the arguments
#   (if the selected element is a function, evaluate it with no parameter)
import datetime
import random
import sys

from .Lexicon import currentLanguage
from .Terminal import Terminal
from .Phrase import Phrase
from .Dependent import Dependent

def oneOf(*elems):
    if len(elems) == 1:
        if isinstance(elems[0], list):
            e = random.choice(elems[0])
        else:
            e = elems[0]
    else:
        e = random.choice(elems)
    return e() if callable(e) else e


# Flag for quoting out of vocabulary tokens (not yet taken into account)
# quoteOOV=False;
# def setQuoteOOV(qOOV):
#     global quoteOOV
#     quoteOOV=qOOV

# create expression from a JSON structure
def fromJSON(json, lang=None):
    if isinstance(json, dict):
        if "lang" in json:
            if json["lang"] == "en":
                lang = "en"
            elif json["lang"] == "fr":
                lang = "fr"
            else:
                print("FromJSON: lang should be 'en' or 'fr', not " + json["lang"] + " 'en' will be used",
                      file=sys.stderr)
                lang = "en"
        lang1 = lang if lang is not None else currentLanguage()
        if "phrase" in json:
            constType = json["phrase"]
            if constType in ['NP', 'AP', 'AdvP', 'VP', 'PP', 'CP', 'S', 'SP']:
                return Phrase.fromJSON(constType, json, lang1)
            else:
                print("fromJSON: unknown Phrase type:" + constType, file=sys.stderr)
        elif "dependent" in json:
            constType = json["dependent"]
            if constType in ["root", "det", "subj", "comp", "mod", "compObj", "compObl", "coord"]:
                return Dependent.fromJSON(constType, json, lang1)
            else:
                print("fromJSON: unknown Dependent type:" + constType, file=sys.stderr)
        elif "terminal" in json:
            constType = json["terminal"]
            if constType in ['N', 'A', 'Pro', 'D', 'Adv', 'V', 'P', 'C', 'DT', 'NO', 'Q']:
                return Terminal.fromJSON(constType, json, lang1)
            else:
                print("fromJSON: unknown Terminal type:" + constType, file=sys.stderr)
    else:
        print("fromJSON: object expected, but found " + type(json).__name__ + ":" + repr(json), file=sys.stderr)


# useful variables for using expressions written originally for the javascript version
false = False
true = True
null = None

# version and date informations
pyRealB_version = "2.1"
pyRealB_dateCreated = datetime.datetime.today()



import json,sys, os

datadir=os.path.abspath(os.path.join(os.path.dirname(__file__),'data')) 

class Lexicon(object):
    '''
    Keep information about English and French lexica and rules 
    '''
    
    def __init__(self):
        ''' load English and French lexica and rules'''
        
        self.lexicon={
            "en":json.load(open(os.path.join(datadir,"lexicon-en.json"),encoding="utf-8")),
            "fr":json.load(open(os.path.join(datadir,"lexicon-fr.json"),encoding="utf-8")),
        }
        self.rules={
            "en":json.load(open(os.path.join(datadir,"rules-en.json"),encoding="utf-8")),
            "fr":json.load(open(os.path.join(datadir,"rules-fr.json"),encoding="utf-8")),
        }
        self.lang="en"
        
    def getLexicalInfo(self,lemma):
        getLexicon=self.getLexicon()
        if lemma in getLexicon:
            return getLexicon[lemma]
        else:
            return None
    
    def getLexicon(self):
        return self.lexicon[self.lang]
    
    def getRules(self):
        return self.rules[self.lang]
        
__lexicon = Lexicon()

def currentLanguage():
    return __lexicon.lang

def loadEn(trace=False):
    __lexicon.lang="en"
    if trace: print("English lexicon and rules loaded",file=sys.stderr)

def loadFr(trace=False):
    __lexicon.lang="fr"
    if trace: print("Règles et lexique français chargés",file=sys.stderr)

# add to lexicon and return the updated object
#     to remove from lexicon (give None as newInfos)
def addToLexicon(lemma,newInfos=None,lang=None):
    lexicon=getLexicon(lang)
    if isinstance(lemma,dict): # convenient when called with a single JSON object as shown in the IDE
        item=list(lemma.items())[0]
        newInfos=item[1]
        lemma=item[0]
    elif newInfos is None: # remove entry
        if lemma in lexicon:
            del lexicon[lemma]
        return
    if lemma in lexicon:
        lexicon[lemma].update(newInfos)
    else:
        lexicon[lemma]=newInfos
    return lexicon[lemma]

# update current lexicon by "merging" the pyrealb lexicon with the current one
#     i.e. adding pyrealb key-value pairs and replacing existing key-value pairs with the pyrealb one
#     newLexicon is a single object with the "correct" structure
def updateLexicon(newLexicon,lang=None):
    lexicon = getLexicon(lang)
    lexicon.update(newLexicon)

# get lemma from lexicon (useful for debugging )
def getLemma(lemma,lang=None):
    lexicon=getLexicon(lang)
    return lexicon[lemma] if lemma in lexicon else None

# return the current lexicon    
def getLexicon(lang=None):
    if lang is not None:
        __lexicon.lang=lang
    return __lexicon.getLexicon()

# return the current rules
def getRules(lang=None):
    if lang is not None:
        __lexicon.lang=lang
    return __lexicon.getRules()



from pyrealb.Lexicon import *
from pyrealb.Terminal import *
from pyrealb.Phrase import *
from pyrealb.Dependent import *
from pyrealb.utils import *
from pyrealb.Warning import *

__all__ = ['Constituent',
     'A', 'Adv', 'C', 'D', 'DT', 'N', 'NO', 'P', 'Pro', 'Q', 'V', 'Terminal',           # from Terminal
     'AP',  'AdvP',  'CP', 'NP', 'PP',  'VP', 'S', 'SP', 'Phrase',                      # from Phrase
     'root', 'subj', 'det', 'mod', 'comp', 'compObj', 'compObl', 'coord', 'Dependent',  # from Dependent
     'currentLanguage', 'addToLexicon', 'getLemma', 'loadEn', 'loadFr',                 # from Lexicon
     'fromJSON', 'oneOf', 'false', 'true', 'null', 'pyRealB_version',                   # from utils
     'test_warnings'                                                                    # from Warning
 ]



from .Constituent import Constituent
from .Terminal import Terminal, N, A, Pro, D, Adv, V, P, C, DT, NO, Q
from .Phrase import Phrase, prepositionsList
from .Lexicon import getLexicon, getRules, currentLanguage
import sys

class Dependent(Constituent):
    def __init__(self, params, deprel, lang=None):
        super().__init__(deprel)
        if lang is None:
            self.lang = currentLanguage()
        elif lang in ["en", "fr"]:
            self.lang = lang
        else:
            self.lang = currentLanguage()
            self.warn("bad language", self.lang)
        self.terminal = Q("*terminal") # dummy terminal so that error messages can be issued
        if len(params)==0:
            self.warn("Dependent without params")
        params=list(params)
        if isinstance(params[0],Terminal):
            self.terminal=params.pop(0)
        else:
            self.warn("Dependent needs Terminal",type(params[0]).__name__)
            params.pop(0)
        self.terminal.parentConst=self
        if hasattr(self.terminal,"peng"):
            self.peng=self.terminal.peng
        if self.terminal.isA("V"):
            self.taux=self.terminal.taux
        params = [e for e in params if e != None]  # ignore None dependents
        # list of dependents to create the source of the parameters at the time of the call
        # self can be different from the dependents lists because of structure modifications
        self.dependentsSource = []
        self.dependents = []
        # add all dependents except the last to the list of dependents
        for i in range(0, len(params) - 1):
            d = params[i]
            if isinstance(d, Dependent):
                self.addDependent(d)
                self.dependentsSource.append(d)
            else:
                self.warn("bad Dependent", NO(i + 1).dOpt({"ord":True}), type(d).__name__+str(d))
        if len(params) > 0:
            # terminate the list with add which does other checks on the final list
            self.add(params[-1], None, True)

    def addDependent(self,dependent,position=None):
        if isinstance(dependent,Dependent):
            dependent.parentConst=self
            # add it to the list of dependents
            if position==None:
                self.dependents.append(dependent)
            elif isinstance(position,int) and position <= len(self.dependents) and position>=0:
                self.dependents.insert(position,dependent)
            else:
                self.warn("bad position", position, len(self.dependents))
        else:
            self.warn("bad Dependent", NO(position + 1).dOpt({"ord": True}), type(dependent).__name__)
        return self

    # remove a Dependent from this Dependent and return it
    def removeDependent(self, position):
        if isinstance(position, int) and 0 <= position < len(self.dependents):
            dependent = self.dependents.pop(position)
            dependent.parentConst = None
            return dependent
        return self.warn("bad position", position, len(self.dependents))

    def changeDeprel(self,newdep):
        if self.isA("coord"):
            for d in self.dependents:
                d.constType=newdep
        else:
            self.constType=newdep
        return self

    # find index of one of deprels taking into account possible coord starting from position start
    # returns -1 when not found
    def findIndex(self,test,start=0):
        for i in range(start,len(self.dependents)):
            d=self.dependents[i]
            if d.isA("coord") and len(d.dependents) and test(d.dependents[0]): return i
            if test(d):return i
        return -1

    def add(self,dependent,position=None,prog=None):
        if not isinstance(dependent,Dependent):
            return self.warn("bad Dependent","dernier" if self.isFr() else "last",type(dependent).__name__)
        if prog is None : # real call .add
            self.optSource+=f'.add({dependent.toSource()}{"" if position is None else (","+str(position))})'
        else:
            self.dependentsSource.append(dependent)
        self.addDependent(dependent,position)
        self.linkProperties()
        return self

    def me(self):
        args=[self.terminal]
        if hasattr(self,"dependents"):
            args+=self.dependents
        return f'{self.constType}({",".join([e.me() for e in args])})'

    def linkProperties(self):
        if len(self.dependents)==0:return self
        # loop over children to set the peng and taux to their head or subject
        # so that once a value is changed this change will be propagated correctly...
        headTerm=self.terminal
        if self.isA("coord"):
            # must create self.peng already because it might be used in the current dependents (for adjectives, attributes...)
            # the information will be computed at realization time (see Dependent.coordReal)
            self.peng={}
        for d in self.dependents:
            depTerm=d.terminal
            deprel=d.constType
            if deprel=="subj":
                if headTerm.isA("V"):
                    headTerm.peng=d.peng
            elif deprel=="det":
                if depTerm.isA("D"):
                    if hasattr(self,"peng"):
                        if hasattr(depTerm,"peng") and hasattr(depTerm.peng,"pe"): # save person (for possessives)
                            pe=depTerm.peng["pe"]
                            depTerm.peng=self.peng
                            depTerm.peng["pe"]=pe
                        else: #some strange determiner construct do not have peng
                            depTerm.peng=self.peng
                elif depTerm.isA("NO"):
                    headTerm.peng["n"]=depTerm.grammaticalNumber()
                    # gender agreement between a French numbeer and subject
                    depTerm.peng["g"]=headTerm.peng["g"]
                elif depTerm.isA("P") and depTerm.lemma=="de": # HACK: deal with specific case : det(P("de"),mod(D(...)))
                    if len(d.dependents)==1 and d.dependents[0].isA("mod") and d.dependents[0].terminal.isA("D"):
                        d.dependents[0].terminal.peng = self.peng
            elif deprel=="mod" or deprel=="comp":
                if depTerm.isA("A"):
                    depTerm.peng=self.peng
                    # check for an attribute of a copula with an adjective
                    if self.isFr() and headTerm.lemma in ["être", "paraître", "sembler", "devenir", "rester"]:
                        iSubj=self.findIndex(lambda d0:d0.isA("subj") and d0.terminal.isA("N"))
                        if iSubj>=0: depTerm.peng=self.dependents[iSubj].peng
                elif depTerm.isA("V"):
                    # set agreement between the subject of a subordinate or the object of a relative subordinate
                    iRel=d.findIndex(lambda d0:d0.isOneOf(["subj","comp","mod"]) and
                                                  d0.terminal.isA("Pro") and
                                                  d0.terminal.lemma in ["qui","que","who","that"])
                    if iRel>=0:
                        rel = d.dependents[iRel].constType
                        if rel=="subj": # verb agrees with this subject
                            depTerm.peng=self.peng
                        elif self.isFr():
                            depTerm.cod=headTerm
                    # check for past participle in French that should agree with the head
                    if self.isFr() and depTerm.getProp("t")=="pp":
                        depTerm.peng=self.peng
            elif deprel=="root":
                self.error("An internal root was found")
            elif deprel=="coord":
                if len(d.dependents)>0:
                    firstDep=d.dependents[0]
                    if firstDep.isA("subj"):
                        headTerm.peng=d.peng
                    elif firstDep.isOneOf(["mod","comp"]) and firstDep.terminal.isOneOf(["V","A"]):
                        # consider this as coordination of verb sharing a subject (the current root)
                        # or coordination of adjectives
                        for d0 in d.dependents:
                            d0.peng=self.peng
                            d0.terminal.peng=self.peng
            else:
                self.error("Strange dependent:"+d.constType)

    def setLemma(self,_lemma,_terminalTyp):
        self.error("***: should never happen: setLemma: called on a Dependent")
        return self

    def findGenderNumberPerson(self,andCombination):
        g="f"
        n="s"
        pe=3
        nb=0
        for i in range(0,len(self.dependents)):
            e=self.dependents[i].terminal
            if e.isOneOf(["NP","N","Pro","Q"]):
                nb += 1
                propG=e.getProp("g")
                if propG=="m" or propG=="x" or e.isA("Q"):g="m" # masculine if gender is unspecified
                if e.getProp("n")=="p":n="p"
                propPe=e.getProp("pe")
                if propPe != None and propPe<pe:pe=propPe
        if nb==0: g="m"
        elif nb>1 and n=="s" and andCombination:n="p"
        return {"g":g,"n":n,"pe":pe}

    # Dependency structure modification but that must be called in the context of the parentConst
    # because the pronoun depends on the role of the N in the sentence
    #         and its position can also change relatively to the verb
    def pronominalize_fr(self):
        mySelf = self
        if self.isA("subj"):
            if not self.terminal.isOneOf(["N", "Pro"]):
                return self.warn("no appropriate pronoun")
            pro = self.getTonicPro("nom")
        elif self.isOneOf(["comp", "mod"]) and self.terminal.isA("P"):
            prep = self.terminal.lemma
            if len(self.dependents) == 1 and self.dependents[0].isOneOf(["comp", "mod"]):
                if self.dependents[0].terminal.isA("N"):
                    n = self.dependents[0].terminal
                    if prep == "à":
                        pro = n.getTonicPro("dat")
                    elif prep == "de":
                        pro = Pro("en", "fr").c("dat")
                    elif prep in ["sur", "vers", "dans"]:
                        pro = Pro("y", "fr").c("dat")
                    else:  # change only the N keeping the P intact
                        pro = self.getTonicPro("nom")
                        mySelf = self.dependents[0]
                else:
                    return self.warn("no appropriate pronoun")
        else:
            pro = self.getTonicPro("acc")
            if self.parentConst != None and self.parentConst.terminal.isA("V"):  # consider that it is direct complement
                self.parentConst.terminal.cod = self  # indicate that self is a COD
        # replace the original with the pronoun
        pro.parentConst = self
        if hasattr(mySelf,"peng"):
            pro.peng = mySelf.peng
        mySelf.terminal = pro
        mySelf.dependents = []
        mySelf.dependentsSource = []

    # Pronominalization in English only applies to a N (self is checked before the call)
    #  and does not need reorganisation of the sentence
    #  Does not currently deal with "Give the book to her." that {c|sh}ould be "Give her the book."
    def pronominalize_en(self):
        self.props["pe"] = 3  # ensure that pronominalization of a noun is 3rd person
        if self.isA("subj"):  # is it a subject
            pro = self.getTonicPro("nom")
        else:
            pro = self.getTonicPro("acc")  # is direct complement
        pro.peng = self.peng
        pro.parentConst = self
        self.terminal = pro
        self.dependents = []
        self.dependentsSource = []

    # check if any child should be pronominalized
    # self must be done in the context of the parent, because some dependents might be changed
    def pronominalizeChildren(self):
        for d in self.dependents:
            if "pro" in d.props and not d.terminal.isA("Pro"):  # it can happen that a Pro has property "pro" set within the same expression
                if d.isFr():
                    d.pronominalize_fr()
                else:
                    if d.terminal.isA("N"):
                        d.pronominalize_en()
                    else:
                        self.warn("no appropriate pronoun")

    def passivate(self):
        subj=None # original subject and object, they are swapped below...
        obj=None
        # find the subject
        if self.terminal.isA("V"):
            subjIdx=self.findIndex(lambda d:d.isA("subj"))
            if subjIdx>=0:
                subj=self.dependents[subjIdx]
                if subj.terminal.isA("Pro"):
                    # as self pronoun will be preceded by "par" or "by", the "bare" tonic form is needed
                    # to which we assign the original person, number, gender
                    subj.terminal=subj.terminal.getTonicPro().g(subj.getProp("g"))\
                                            .n(subj.getProp("n")).pe(subj.getProp("pe"))
            else:
                subj=None
            # find direct object (first N or Pro of a comp) from dependents
            objIdx=self.findIndex(lambda d: d.isA("comp") and d.terminal.isOneOf(["N","Pro"]))
            if objIdx>=0:
                obj=self.dependents[objIdx]
                if obj.terminal.isA("Pro"):
                    obj.terminal=obj.terminal.getTonicPro("nom")
                elif obj.terminal.isA("N") and "pro" in obj.props:
                    obj=obj.getTonicPro("nom")
                # swap subject and object by changing their relation name !!!
                # HACK: obj is the new subject
                obj.changeDeprel("subj")
                # make the verb agrees with the new subject (in English only, French is dealt below)
                if self.isEn():
                    self.terminal.peng=obj.peng
                if subj is not None:   # the original subject is now the indirect object
                    subj.changeDeprel("mod")
                    self.removeDependent(subjIdx)
                    self.addDependent(comp(P("par" if self.isFr() else "by",self.lang),subj))
            elif subj is not None: # no object, but with a subject
                #create a dummy subject with a "il"/"it"
                obj=Pro("lui" if self.isFr() else "it",self.lang).c("nom") #HACK: obj is the new subject
                # add new subject at the front of the sentence
                self.addPre(obj)
                self.peng=obj.peng
                 # add original subject after the verb to serve as an object
                self.addDependent(comp(P("par" if self.isFr() else "by",self.lang),subj))
            if self.isFr():
                # do self only for French because in English self is done by processTyp_en
                # change verbe into an "être" auxiliary and make it agree with the newSubj
                # force person to be 3rd (number and tense will come from the new subject)
                verbe=self.terminal.lemma
                self.terminal.setLemma("être")
                self.terminal.pe(3)
                if self.getProp("t")=="ip":
                    self.t("s") # set subjonctive present tense for an imperativ
                pp = V(verbe,"fr").t("pp")
                if obj is not None: # self can be undefined when a subject is Q or missing
                    self.terminal.peng=obj.peng
                    pp.peng=obj.peng
                # insert the pp before the comp, so that it appears immediately after the verb
                #  calling addPost(pp) would not put it at the right place
                compIdx=self.findIndex(lambda d:d.isA("comp"))
                if compIdx==-1: compIdx=0
                self.addDependent(Dependent([pp],"*post*"),compIdx)
        else:
            return self.warn("not found","V","contexte passif" if self.isFr() else "passive context")

    def processV(self, types, key, action):
        if self.isA("coord"):
            for d in self.dependents:
                d.processV(types, action)
        elif self.terminal.isA("V"):
            if key in types and types[key] is not False:
                action(self, types[key])

    # add special internal dependencies (used only during realization)
    def addPre(self, terminals, position=None):
        if isinstance(terminals, Terminal):
            self.addDependent(Dependent([terminals], "*pre*"), position)
            return self
        for terminal in terminals:
            self.addDependent(Dependent([terminal], "*pre*"), position)
        return self

    def addPost(self, terminals):
        if isinstance(terminals, Terminal):
            self.addDependent(Dependent([terminals], "*post*"), 0)
            return self
        terminals.reverse()  # must add them in reverse order because of position 0
        for terminal in terminals:
            self.addDependent(Dependent([terminal], "*post*"), 0)
        return self

    def processTyp_fr(self, types):
        def progF(deprel, v):
            # insert "en train","de" (separate so that élision can be done...)
            # HACK::but do it BEFORE the pronouns created by .pro()
            origLemma = deprel.terminal.lemma
            deprel.terminal.setLemma(
                "être")  # change verb, but keep person, number and tense properties of the original...
            deprel.addPost([Q("en train"), Q("de"), V(origLemma).t("b")])
            deprel.terminal.isProg = v
        self.processV(types, "prog", progF)

        def modF(deprel, mod):
            rules = getRules()
            origLemma = deprel.terminal.lemma
            for key in rules["verb_option"]["modalityVerb"]:
                if (key.startswith(mod)):
                    deprel.terminal.setLemma(rules["verb_option"]["modalityVerb"][key])
                    break
            deprel.terminal.isMod = True
            newV = V(origLemma).t("b")
            if hasattr(deprel.terminal, "isProg"):  # copy progressive from original verb...
                newV.isProg = deprel.terminal.isProg
                del deprel.terminal.isProg
            deprel.addPost(newV)
        self.processV(types, "mod", modF)

        def negF(deprel, neg):
            if neg is True: neg="pas"
            deprel.terminal.neg2 = neg  # HACK: to be used when conjugating at the realization time
        self.processV(types, "neg", negF)

    def processTyp_en(self, types):
        # replace current verb with the list new words
        #  TODO: take into account the fact that there might be already a verb with modals...
        if ("contr" in types and types["contr"] is not False):
            # necessary because we want the negation to be contracted within the VP before the S or SP
            self.contraction = True
        words = Phrase.affixHopping(self.terminal, self.getProp("t"), getRules()["compound"], types)
        # the new root should be the last verb
        last = words.pop()
        if last.isA("Pro") and last.lemma == "myself":  # skip possible "myself" for reflexive verb
            self.addPost(last)
            last = words.pop()
        self.terminal = last
        self.addPre(words)

    def moveAuxToFront(self):
        auxIdx = self.findIndex(lambda d: d.isA("*pre*"))  # added by affixHopping
        if auxIdx >= 0:
            aux = self.dependents[auxIdx].terminal
            self.removeDependent(auxIdx)
            self.addPre(aux, 0)  # put auxiliary before

    def invertSubject(self):
        # in French : use inversion rule which is quite "delicate"
        # rules from https:#francais.lingolia.com/fr/grammaire/la-phrase/la-phrase-interrogative
        # if subject is a pronoun, invert and add "-t-" or "-"
        # if subject is a noun, the subject stays but add a new pronoun
        subjIdx = self.findIndex(lambda d: d.isA("subj"))
        if subjIdx >= 0:
            subj = self.dependents[subjIdx].terminal
            if subj.isA("Pro"):
                pro = self.removeDependent(subjIdx).terminal  # remove subject
            elif subj.isA("C"):
                pro = Pro("moi", "fr").c("nom").g("m").n("p").pe(3)  # create a "standard" pronoun, to be patched by cpReal
                subj.pronoun = pro  # add a flag to be processed by cpReal
            else:
                pro = Pro("moi", "fr").g(subj.getProp("g")).n(subj.getProp("n")).pe(3).c("nom")  # create a pronoun
            if self.terminal.isA("V"):
                self.addPost(pro)
                self.terminal.lier()

    def processTypInt(self,int_):
        sentenceTypeInt = getRules()["sentence_type"]["int"]
        intPrefix = sentenceTypeInt["prefix"]
        prefix = None  # to be filled later
        pp = None
        if int_ in ["yon", "how", "why", "muc"]:
            if self.isEn(): self.moveAuxToFront()
            else: self.invertSubject()
            prefix = intPrefix[int_]
        # remove a part of the sentence
        elif int_ in ["wos", "was"]:  # remove subject
            subjIdx = self.findIndex(lambda d: d.isA("subj"))
            if subjIdx >= 0:
                # insure that the verb at the third person singular,
                # because now the subject has been removed
                self.terminal.setProp("n", "s")
                self.terminal.setProp("pe", 3)
                self.removeDependent(subjIdx)
            prefix = intPrefix[int_]
        elif int_ in ["wod", "wad"]:  # remove direct object (first comp starting with N)
            for i in range(0, len(self.dependents)):
                d = self.dependents[i]
                if d.isA("comp") and d.terminal.isA("N"):
                    self.removeDependent(i)
            if self.isFr():  # check for passive subject starting with par
                parIdx = self.findIndex(
                    lambda d: d.isA("comp") and d.terminal.isA("P") and d.terminal.lemma == "par")
                if parIdx >= 0:
                    pp = self.dependents[parIdx].terminal
                    self.removeDependent(parIdx)  # remove the passive subject
            prefix = intPrefix[int_]
            if self.isEn():
                self.moveAuxToFront()
            else:
                self.invertSubject()
        elif int_ in ["woi", "wai", "whe", "whn"]:  # remove indirect object first comp or mod with a P as terminal
            remove = False
            prefix = intPrefix[int_]  # get default prefix
            for i in range(0, len(self.dependents)):
                d = self.dependents[i]
                if d.isOneOf(["comp", "mod"]) and d.terminal.isA("P"):
                    # try to find a more appropriate prefix by looking at preposition in the structure
                    prep = d.terminal.lemma
                    preps = prepositionsList[self.lang]
                    remove = False
                    if int == "whe":
                        if preps["whe"].has(prep): remove = True
                    elif int == "whn":
                        if preps["whn"].has(prep): remove = True
                    elif preps["all"].has(prep):  # "woi" | "wai"
                        # add the preposition in front of the prefix (should be in the table...)
                        prefix = prep + " " + (("whom" if int == "woi" else "what") if self.isEn()
                                               else ("qui" if int == "woi" else "quoi"))
                        remove = True
                    if remove:
                        self.removeDependent(i)
                        break
            if self.isEn():
                self.moveAuxToFront()
            else:
                self.invertSubject()
        else:
            self.warn("not implemented", "int:" + int_)
        if self.isFr() or (int_ != "yon"):   # add the interrogative prefix
            self.addPre(Q(prefix), 0)
        if pp != None:  # add "par" in front of some French passive interrogative
            self.addPre(pp, 0)
        if int_ == "wad":  # replace "que" by "quoi" for French passive wad
            self.dependents[1].terminal.lemma = "quoi"
        self.a(sentenceTypeInt["punctuation"], True)

    def processTyp(self, types):
        # pp=None # flag for possible pp removal for French wod or wad
        if "pas" in types and types["pas"] is not False:
            self.passivate()
        if self.isFr():
            if "contr" in types and types["contr"] is not False:
                self.warn("no French contraction")
            self.processTyp_fr(types)
        else:
            self.processTyp_en(types)
        if "int" in types and types["int"] is not False:
            self.processTypInt(types["int"])
        if "exc" in types and types["exc"] is True:
            self.a(getRules()["sentence_type"]["exc"]["punctuation"], True)
        return self

    def coordReal(self):
        res=[]
        # realize coordinated Dependents by adding ',' between all dependents except for the last
        # no check is done on the terminal, so
        #    if the terminal is Q(",") then all dependents are separated by a ","
        #    Q("and,") deal with the Oxford comma (i.e. a comma after all dependents even the last)
        last=len(self.dependents)-1
        if last==-1: return []
        if last==0: # coordination with only one element, ignore coordinate
            dep = self.dependents[0]
            res = dep.real()
            self.setProp("g",dep.getProp("g"))
            self.setProp("n",dep.getProp("n"))
            pe=dep.getProp("pe")
            self.setProp("pe",pe if pe!=None else 3)
            return self.doFormat(res) # process format for the CP
        # check that all dependents use the same deprel
        deprel=self.dependents[0].constType
        for j in range(0,last): #insert comma after each element
            dj=self.dependents[j]
            if not dj.isA(deprel):
                self.warn("inconsistent dependents within a coord",deprel,dj.constType)
            if j<last-1: dj.props["a"]=[","]
            res.extend(dj.real())
        # insert realisation of the terminal before last...
        res.extend(self.terminal.real())
        if not self.dependents[last].isA(deprel):
            self.warn("inconsistent dependents within a coord",deprel,self.dependents[last].constType)
        # insert last element
        res.extend(self.dependents[last].real())
        # compute the combined gender and number of the coordination once children have been realized
        selfCoord=self.terminal
        if selfCoord.isA("C"):
            andC="et" if self.isFr() else "and"
            gn=self.findGenderNumberPerson(selfCoord.lemma==andC)
            self.setProp("g",gn["g"])
            self.setProp("n",gn["n"])
            self.setProp("pe",gn["pe"])
            # for an inserted pronoun, we must override its existing properties...
            if  hasattr(selfCoord,"pronoun"):
                selfCoord.pronoun.peng=gn
                selfCoord.pronoun.props["g"]=gn["g"]
                selfCoord.pronoun.props["n"]=gn["n"]
                selfCoord.pronoun.props["pe"]=gn["pe"]
        elif not selfCoord.isA("Q"):
            # in some cases the coordination can be a quoted string (possibly empty)
            self.warn("bad parameter","C",selfCoord.constType)
        return res

    def depPosition(self):
        myPos = self.props["pos"] if "pos" in self.props else None
        if myPos=="pre": return "pre"
        pos="post" # default is post
        if self.isOneOf(["subj","det","*pre*"]) and myPos!="post":
            # subject and det are always pre except when specified
            pos="pre"
        elif self.isA("mod") and self.terminal.isA("A") and self.parentConst.terminal.isA("N"):
            # check adjective position with respect to a noun
            if self.isFr():
                pos=self.terminal.props["pos"] if "pos" in self.terminal.props else "post"
            else:
                pos="pre" # all English adjectives are pre
        elif self.isA("coord"):
            if myPos is not None:
                pos=myPos
            elif len(self.dependents)>0:
                pos=self.dependents[0].depPosition() #take the position of the first element of the coordination
        return pos

    def real(self):
        self.pronominalizeChildren()
        if "typ" in self.props:
            self.processTyp(self.props["typ"])
        # realize coordinations before anything elso to compute their final number and person
        for d in self.dependents:
            if d.isA("coord"):d.tokens=d.coordReal()
        before=[]
        after=[]
        # realize and order them by gathering the dependents that should appear before and after the terminal
        for d in self.dependents:
            if d.depPosition()=="pre":
                before += d.tokens if d.isA("coord") else d.real()
            else:
                after += d.tokens if d.isA("coord") else d.real()
        res=[*before,*self.terminal.real(),*after]
        return self.doFormat(res)

    # recreate a jsRealB expression
    # if indent is >=0 create an indented pretty-print (call it with 0 at the root)
    def toSource(self, indent=-1):
        if indent >= 0:
            indent = indent + len(self.constType) + 1
            sep = ",\n" + " " * indent
        else:
            sep = ","
        # create source of children
        deps=[self.terminal.toSource(indent)]+[e.toSource(indent) for e in self.dependentsSource]
        return f'{self.constType}({sep.join(deps)})' + super().toSource()

    def toJSON(self):
        res = {"dependent": self.constType,
               "terminal":self.terminal.toJSON()}
        if hasattr(self,"dependents"):
            res["dependents"]=[e.toJSON() for e in self.dependents]
        if len(self.props) > 0:  # do not output empty props
            res["props"] = self.props
        if self.parentConst is None or self.lang != self.parentConst.lang:  # only indicate when language changes
            res["lang"] = self.lang
        return res

    @classmethod
    def fromJSON(self, constType, json, lang):
        from .utils import fromJSON
        if "terminal" not in json:
            print("Dependent.fromJSON: no terminal found in Dependent:"+str(json))
        else:
            args=[fromJSON(json["terminal"],lang)]
            if "dependents" in json:
                if isinstance(json["dependents"], list):
                    args.extend([fromJSON(e, lang) for e in json["dependents"]])
                else:
                    print("Dependent.fromJSON dependents should be a list:" + str(json["elements"]))
            return Dependent(args, constType, lang).setJSONprops(json)

#  create Dependent instances
def root(*params): return Dependent(params,"root")
def subj(*params): return Dependent(params,"subj")
def det(*params): return Dependent(params,"det")
def mod(*params): return Dependent(params,"mod")

def comp(*params): return Dependent(params,"comp")
def compObj(*params): return Dependent(params,"comp")
def compObl(*params): return Dependent(params,"comp")

def coord(*params): return Dependent(params,"coord")




import re,datetime,sys

from .Lexicon import getLexicon, getLemma, getRules, currentLanguage

defaultProps = {"en":{"g":"n","n":"s","pe":3,"t":"p"},             # language dependent default properties
                "fr":{"g":"m","n":"s","pe":3,"t":"p","aux":"av"}}

quoteOOV=False # TODO: make this settable globally
optionListMethods = ('a', 'b', 'ba', 'en')

deprels = ["root","subj","det","mod","comp","coord"] # list of implemented dependency relations

# regex for matching the first word in a generated string (ouch!!! it is quite subtle...) 
#  match index:
#     1-possible non-word chars and optional html tags
#     2-the real word 
#     3-the rest after the word  
sepWordREen=re.compile(r"((?:[^<\w'-]*(?:<[^>]+>)?)*)([\w'-]+)?(.*)")
# same as sepWordREen but the [\w] class is extended with French accented letters and cedilla
sepWordREfr=re.compile(r"((?:[^<\wàâéèêëîïôöùüç'-]*(?:<[^>]+>)?)*)([\wàâéèêëîïôöùüç'-]+)?(.*)",re.I)

# tables des positions des clitiques en français, tirées de
#    Choi-Jonin (I.) & Lagae (V.), 2016, « Les pronoms personnels clitiques », in Encyclopédie Grammaticale du Français,
#    en ligne : http:#encyclogram.fr

#  section 3.1.1. (http:#encyclogram.fr/notx/006/006_Notice.php#tit31)

proclitiqueOrdre = {  # page 11 du PDF
    # premier pronom que nous ignorons pour les besoins de cette application
    # "je":1, "tu":1, "il":1, "elle":1, "on":1, "on":1, "nous":1, "vous":1, "vous":1, "ils":1, "elle":1,
    "ne": 2,
    "me": 3, "te": 3, "se": 3, "nous": 3, "vous": 3,
    "le": 4, "la": 4, "les": 4,
    "lui": 5, "leur": 5,
    "y": 6,
    "en": 7,
    "*verbe*": 8,
    "pas": 9,  # S'applique aussi aux autre négations... plus, guère
}

proclitiqueOrdreImperatifNeg = {  # page 14 du PDF
    "ne": 1,
    "me": 2, "te": 2, "nous": 2, "vous": 2,
    "le": 3, "la": 3, "les": 3,
    "lui": 4, "leur": 4,
    "y": 5,
    "en": 6,
    "*verbe*": 7,
    "pas": 8,  # S'applique aussi aux autre négations... plus, guère
}

proclitiqueOrdreImperatifPos = {  # page 15 du PDF
    "*verbe*": 1,
    "le": 2, "la": 2, "les": 2,
    "lui": 3, "leur": 3,
    "me": 4, "te": 4, "nous": 2, "vous": 2,
    "y": 5,
    "en": 6,
}

proclitiqueOrdreInfinitif = {  # page 17 du PDF
    "ne": 1,
    "pas": 2,  # S'applique aussi aux autre négations... plus, guère, jamais
    "me": 3, "te": 3, "se": 3, "nous": 3, "vous": 3,
    "le": 4, "la": 4, "les": 4,
    "lui": 5, "leur": 5,
    "y": 6,
    "en": 7,
    "*verbe*": 8,
}

modalityVerbs = ["vouloir", "devoir", "pouvoir"]


class Constituent():
    def __init__(self,constType):
        self.constType=constType
        # initialize other params
        self.parentConst=None
        self.props = {}
        self.realization=None 
        self.optSource=""     # string corresponding to the calls to the options
    
    def error(self,mess):
        raise Exception ("Internal error: this should never have happened,sorry!\n"+self.me()+"::"+mess)
    
    def isA(self,aType):
        return self.constType==aType
    
    def isOneOf(self,types):
        return self.constType in types
    
    def isFr(self):
        return self.lang=="fr"
    
    def isEn(self):
        return self.lang=="en"
    
    # def __pos__(self): # prefix + :: useful trick for testing top-level expression realization
    #     print(str(self))
    
    # get/set the value of a property by first checking the special shared properties
    def getProp(self,propName):
        if propName in self.props:
            return self.props[propName] 
        if propName in ["pe","n","g"]:
            if not hasattr(self, "peng") or self.peng==None: return None
            return self.peng[propName] if propName in self.peng else None
        if propName in ["t","aux"]:
            if not hasattr(self,"taux") or self.taux==None: return None
            return self.taux[propName] if propName in self.taux else None 
        return None
    
    def setProp(self,propName,val):
        if propName in ["pe","n","g"] and hasattr(self,"peng") and self.peng is not None:
            self.peng[propName]=val
        if propName in ["t","aux"] and hasattr(self,"taux") and self.taux is not None:
            self.taux[propName]=val
        self.props[propName]=val

    # should be in Terminal.prototype... but here for consistency with three previous definitions
    pengNO=0 # useful for debugging: identifier of peng struct to check proper sharing in the debugger
    tauxNO=0 # useful for debugging: identifier of taux struct to check proper sharing in the debugger
   
    def initProps(self):
        if self.isOneOf(["N","A","D","V","NO","Pro","Q","DT"]):
            self.peng={"pe":defaultProps[self.lang]["pe"],
                       "n" :defaultProps[self.lang]["n"],
                       "g" :defaultProps[self.lang]["g"]}
            Constituent.pengNO+=1
            self.peng["pengNO"]=Constituent.pengNO
            if self.isA("V"):
                self.taux={"t":defaultProps[self.lang]["t"]}
                Constituent.tauxNO+=1
                self.taux["tauxNO"]=Constituent.tauxNO
                if self.isFr():
                    self.taux["aux"]=defaultProps[self.lang]["aux"]
    
    # get a given constituent with a path starting at this
    #   path is a list of node type , or list of node types (an empty string in this list means optional)
    #   returns undefined if any node does not exist on the path
    def getFromPath(self,path):
        if len(path)==0: return self
        current=path.pop(0)
        c=self.getConst(current)
        if c is None:
            if isinstance(current,list) and "" in current and len(path)>0: # optional
                return self.getFromPath(path)
            return None
        return c.getFromPath(path)

    # return a pronoun corresponding to this object 
    # taking into account the current gender, number and person
    #  do not change the current pronoun, if it is already using the tonic form or does not have one (e.g. this)
    # if case_ is not given, return the tonic form else return the corresponding case
    # HACK:: parameter case_ is followed by _ so that it is not displayed as a keyword in the editor
    def getTonicPro(self,case_=None):
        from .Terminal import Pro
        if self.isA("Pro"):
            if ("tn" in self.props or "c" in self.props):
                if case_ is not None:
                    self.props["c"]=case_
                else: # ensure tonic form
                    self.props["tn"]=""
                    if "c" in self.props: del self.props["c"]
            return self
        else: # generate the string corresponding to the tonic form
            pro=Pro("moi" if self.isFr() else "me",self.lang)
            g=self.getProp("g")
            if g!=None: pro.g(g)
            n=self.getProp("n")
            if n!=None: pro.n(n)
            pe=self.getProp("pe")
            if pe!=None: pro.pe(pe)
            if case_ is None:return Pro(pro.realize(),self.lang).tn("")
            return Pro(pro.realize(),self.lang).c(case_)
    
    def getParentLang(self):
        if hasattr(self,"lang"):return self.lang
        if self.parentConst is None: return getLexicon().lang
        return self.parentConst.getParentLang()
                
    def addOptSource(self,optionName,val):
        self.optSource+=f'.{optionName}({repr(val)})'

    ###  genOptionFunc : see makeOptionMethod below
    
    ###  genOptionListFunc: see makeOptionListMethod below
    
    ## HTML tags
    def tag(self,name,attrs=None):
        if attrs is None or len(attrs)==0:
            self.addOptSource("tag",name)
            attrs={}
        else:
            self.optSource+=f'.tag("{name}",{str(attrs)})' # specila case of addOptionSource
        if "tag" not in self.props:self.props["tag"]=[]
        self.props["tag"].append([name,attrs])
        return self
    
    def parseDateString(self,dateS):
        try:  ## parse the string as an iso-like time format 
            m=re.match(r"(\d{4}-\d{2}-\d{2})([T ](\d{2}:\d{2}:\d{2}))?",dateS)
            if m==None:
                self.warn("bad parameter","an isotime format string",dateS)
                return datetime.datetime.today()
            else:
                if m[2]==None:
                    return datetime.datetime(int(m[1][0:4]),int(m[1][5:7]),int(m[1][8:10]))
                else:
                    return datetime.datetime(int(m[1][0:4]),int(m[1][5:7]),int(m[1][8:10]),
                                                hour=int(m[3][0:2]),minute=int(m[3][3:5]),
                                                second=int(m[3][6:8]))
        except ValueError as e:
            self.warn("bad parameter","an isotime format string",dateS+":"+str(e))
            return datetime.datetime.today()

    
    ## date and number options
    def dOpt(self,dOptions):
        self.addOptSource("dOpt",dOptions)
        if not isinstance(dOptions,dict):
            return self.warn("bad application","dOpt","dict",type(dOptions).__name__)
        if self.isA("DT"):
            allowedKeys =["year" , "month" , "date" , "day" , "hour" , "minute" , "second" , "nat", "det", "rtime"]
            if "dOpt" not in self.props:self.props["dOpt"]={}
            for key,val in dOptions.items():
                if key in allowedKeys:
                    if key=="rtime":
                        if isinstance(val,bool):
                            self.props["dOpt"]["rtime"]=datetime.datetime.today() if val else False
                        elif isinstance(val,str):
                            self.props["dOpt"]["rtime"]=self.parseDateString(val)
                        elif isinstance(val,datetime.datetime):
                            self.props["dOpt"]["rtime"]=val
                        else:
                            return self.warn("bad application",'dOpt("rtime")',["bool","str","datetime"],val)
                    elif isinstance(val,bool):
                        self.props["dOpt"][key]=val
                    else:
                        return self.warn("bad application",f'dOpt("{key}")',["bool"],val)
                else:
                    return self.warn("ignored value for option","DT.dOpt",key)
        elif self.isA("NO"):
            allowedKeys = ["mprecision","raw","nat","ord"]
            if "dOpt" not in self.props:self.props["dOpt"]={}
            for key,val in dOptions.items():
                if key in allowedKeys:
                    if key=="mprecision":
                        if isinstance(val,int):
                            self.props["dOpt"]["mprecision"]=val
                        else:
                            return self.warn("bad application","precision",val)
                    elif isinstance(val,bool):
                        self.props["dOpt"][key]=val
                    else:
                        return self.warn("bad application",f'.dOpt("{key}")',"bool",val)
                else:
                    return self.warn("ignored value for option","NO.dOpt",key)
        else:
            return self.warn("bad application",".dOpt",["DT","NO"],self.constType)
        return self

    # number option
    def nat(self,isNat=True):
        self.addOptSource("nat",isNat)
        if self.isOneOf(["DT","NO"]):
            if "dOpt" not in self.props:self.props["dOpt"]={}
            if isNat is None:
                self.props["dOpt"]["nat"]=False
            elif isinstance(isNat,bool):
                self.props["dOpt"]["nat"]=isNat
            else:
                return self.warn("bad application",".nat","bool",isNat)
        else:
            return self.warn("bad application",".nat",["DT","NO"],self.constType)
        return self
    
    def typ(self,types):
        allowedTypes={
              "neg": [False,True],
              "pas": [False,True],
              "prog":[False,True],
              "exc": [False,True],
              "perf":[False,True],
              "refl":[False,True],
              "contr":[False,True],
              "mod": [False,"poss","perm","nece","obli","will"],
              "int": [False,"yon","wos","wod","woi","was","wad","wai","whe","why","whn","how","muc"]
        }
        self.addOptSource("typ",types)
        if not isinstance(types, dict):
            self.warn("ignored value for option",".typ",type(types).__name__+":"+str(types))
        elif not self.isOneOf(["S", "SP", "VP", *deprels]):
            self.warn("bad application", f'.typ("{str(types)}")', ["S", "SP", "VP"], self.constType)
        else: # validate types and keep only the valid ones
            for key,val in types.copy().items():
                if key not in allowedTypes:
                    self.warn("unknown type",key,allowedTypes.keys())
                else:
                    if key=="neg" and self.isFr(): # also accept string as neg value in French
                        if not isinstance(val,(str,bool)):
                            self.warn("ignored value for option",f'.typ("{key}")',val)
                            del types[key]
                    elif val not in allowedTypes[key]:
                        self.warn("ignored value for option",f'.typ("{key}")',val)
                        del types[key]
            self.props["typ"]=types
        return self
                
    def doElisionEn(self,cList):
        # English elision rule only for changing "a" to "an"
        # according to https:#owl.english.purdue.edu/owl/resource/591/1/
        hAnRE=re.compile(r"^(heir|herb|honest|honou?r(able)?|hour)",re.I)
        # https:#www.quora.com/Where-can-I-find-a-list-of-words-that-begin-with-a-vowel-but-use-the-article-a-instead-of-an
        uLikeYouRE=re.compile(r"^(uni.*|ub.*|use.*|usu.*|uv.*)",re.I)
        acronymRE=re.compile(r"^[A-Z]+$")
        # Common Contractions in the English Language taken from :http:#www.everythingenglishblog.com/?p=552
        contractionEnTable={
            "are+not":"aren’t", "can+not":"can’t", "did+not":"didn’t", "do+not":"don’t", "does+not":"doesn’t", 
            "had+not":"hadn’t", "has+not":"hasn’t", "have+not":"haven’t", "is+not":"isn’t", "must+not":"mustn’t", 
            "need+not":"needn’t", "should+not":"shouldn’t", "was+not":"wasn’t", "were+not":"weren’t", 
            "will+not":"won’t", "would+not":"wouldn’t",
            "let+us":"let’s",
            "I+am":"I’m", "I+will":"I’ll", "I+have":"I’ve", "I+had":"I’d", "I+would":"I’d",
            "she+will":"she’ll", "he+is":"he’s", "he+has":"he’s", "she+had":"she’d", "she+would":"she’d",
            "he+will":"he’ll", "she+is":"she’s", "she+has":"she’s", "he+would":"he’d", "he+had":"he’d",
            "you+are":"you’re", "you+will":"you’ll", "you+would":"you’d", "you+had":"you’d", "you+have":"you’ve",
            "we+are":"we’re", "we+will":"we’ll", "we+had":"we’d", "we+would":"we’d", "we+have":"we’ve",
            "they+will":"they’ll", "they+are":"they’re", "they+had":"they’d", "they+would":"they’d", "they+have":"they’ve",
            "it+is":"it’s", "it+will":"it’ll", "it+had":"it’d", "it+would":"it’d",
            "there+will":"there’ll", "there+is":"there’s", "there+has":"there’s", "there+have":"there’ve",
            "that+is":"that’s", "that+had":"that’d", "that+would":"that’d", "that+will":"that’ll"
        } 
        # search for terminal "a" and check if it should be "an" depending on the next word
        last=len(cList)-1
        if last==0: return # do not try to elide a single word
        i=0
        while i<last:
            m1=sepWordREen.match(cList[i].realization)
            if m1 is None or m1.group(2) is None: 
                i+=1
                continue
            m2=sepWordREen.match(cList[i+1].realization)
            if m2 is None or m2.group(2) is None: 
                i+=1
                continue
            # HACK: m1 and m2 save the parts before and after the first word (w1 and w2) which is in m_i.group(2)
            # for a single word 
            w1=m1.group(2)
            w2=m2.group(2)
            if (w1=="a" or w1=="A") and cList[i].isA("D"):
                if (re.match(r"^[aeio]",w2,re.I) or   # starts with a vowel
                    (re.match(r"^u",w2,re.I) and not uLikeYouRE.match(w2) or  # u does not sound like you
                    hAnRE.match(w2) or       # silent h
                    acronymRE.match(w2))):   # is an acronym
                        cList[i].realization=m1.group(1)+w1+"n"+m1.group(3)
                        i+=1                 # skip next word
            elif hasattr(self,"contraction") and self.contraction == True:
                if w1=="cannot": # special case...
                    cList[i].realization=m1.group(1)+"can't"+m1.group(3)
                else:
                    key=w1+"+"+w2
                    if key in contractionEnTable:
                        contr=contractionEnTable[key]
                        # do contraction of first word and remove second word (keeping start and end)
                        cList[i].realization=m1.group(1)+contr+m1.group(3)
                        cList[i+1].realization=m2.group(1)+m2.group(3).strip()
                        i+=1
            i+=1
        
    def doElisionFr(self,cList):
        # Elision rules for French
        # implements the obligatory elision rules of the "Office de la langue française du Québec"
        #    http:#bdl.oqlf.gouv.qc.ca/bdl/gabarit_bdl.asp?Th=2&t1=&id=1737
        # for Euphonie, rules were taken from Antidote (Guide/Phonétique)

        elidableWordFrRE=re.compile(r"^(la|le|je|me|te|se|de|ne|que|puisque|lorsque|jusque|quoique)$",re.I)
        euphonieFrRE=re.compile(r"^(ma|ta|sa|ce|beau|fou|mou|nouveau|vieux)$",re.I)
        euphonieFrTable={"ma":"mon","ta":"ton","sa":"son","ce":"cet",
            "beau":"bel","fou":"fol","mou":"mol","nouveau":"nouvel","vieux":"vieil"}

        contractionFrTable={
            "à+le":"au","à+les":"aux","ça+a":"ç'a",
            "de+le":"du","de+les":"des","de+des":"de","de+autres":"d'autres",
            "des+autres":"d'autres",
            "si+il":"s'il","si+ils":"s'ils"}

        def isElidableFr(realization,lemma,pos):
            # check if realization starts with a vowel
            if re.match(r"^[aeiouyàâéèêëîïôöùü]",realization,re.I): return True
            if re.match(r"^h",realization,re.I):
                #  check for a French "h aspiré" for which no elision should be done
                lexiconInfo=getLemma(lemma if isinstance(lemma,str) else realization) # get the lemma with the right pos
                if lexiconInfo  is None: 
                    lexiconInfo=getLemma(lemma.toLowerCase()) # check with lower case
                    if lexiconInfo  is None: return True       # elide when unknown
                if pos not in lexiconInfo:pos=lexiconInfo.keys()[0] # try the first pos if current not found
                if pos in lexiconInfo and "h" in lexiconInfo[pos] and lexiconInfo[pos]["h"]==1: return False # h aspiré found
                return True
            return False

        contr=None
        last=len(cList)-1
        if last==0:return # do not try to elide a single word
        i=0
        while i < last:
            if i>0 and cList[i-1].getProp("lier")!=None: # ignore if the preceding word is "lié" to this one
                i+=1
                continue
            m1=sepWordREfr.match(cList[i].realization) if cList[i].realization!=None else None
            if m1  is None or m1.group(2) is None: continue
            m2=sepWordREfr.match(cList[i+1].realization) if cList[i].realization!=None else None
            if m2  is None or m2.group(2) is None: continue
            # HACK: m1 and m2 save the parts before and after the first word (w1 and w2) which is in m_i[2]
            # for a single word
            w1=m1.group(2)
            w2=m2.group(2)
            w3NoWords = not re.match(r"^\s*\w",m1.group(3)) # check that the rest of the first word does not start with a word
            if (elidableWordFrRE.match(w1) and isElidableFr(w2,cList[i+1].lemma,cList[i+1].constType) and w3NoWords):
                cList[i].realization=m1[1]+w1[:-1]+"'"+m1[3]
                i+=1
            elif euphonieFrRE.match(w1) and isElidableFr(w2,cList[i+1].lemma,cList[i+1].constType) and  w3NoWords: # euphonie
                if re.match(r"ce",w1,re.I) and re.match(r"(^est$)|(^étai)",w2,re.I):
                    # very special case but very frequent
                    cList[i].realization=m1[1]+w1[:-1]+"'"+m1[3]
                else:
                    cList[i].realization=m1[1]+euphonieFrTable[w1]+m1[3]
                i+=1
            elif (w1+"+"+w2) in contractionFrTable and w3NoWords:
                contr=contractionFrTable[w1+"+"+w2]
                # check if the next word would be elidable, so instead elide it instead of contracting
                # except when the next word is a date which has a "strange" realization
                if (elidableWordFrRE.match(w2) and i+2<=last and not cList[i+1].isA("DT") and
                    isElidableFr(cList[i+2].realization,cList[i+2].lemma,cList[i+2].constType)):
                    cList[i+1].realization=m2[1]+w2[:-1]+"'"+m2[3]
                else: # do contraction of first word and remove second word (keeping start and end)
                    cList[i].realization=m1[1]+contr+m1[3]
                    cList[i+1].realization=m2[1]+m2[3].strip()
                i+=1
            i+=1


    def doFrenchPronounPlacement(self, cList):
        from .Terminal import Adv,Pro
        iDeb = 0
        i = iDeb
        while i < len(cList):
            c = cList[i]
            if c.isA("V") and hasattr(c, "neg2"):
                if hasattr(c, "isMod") or hasattr(c, "isProg"):
                    c.insertReal(cList, Adv(c.neg2, "fr"), i + 1)
                    c.insertReal(cList, Adv("ne", "fr"), i)
                    del c.neg2  # remove negation from the original verb
                    iDeb = i + 3  # skip these in the following loop
                    if hasattr(c, "isProg"): iDeb += 2  # skip "en train","de"
            i += 1
        # gather verb position and pronouns coming after the verb possibly adding a reflexive pronoun
        verbPos = None
        prog = None
        neg2 = None
        pros = []
        i = iDeb
        while i < len(cList):
            c = cList[i]
            if c.isA("V"):
                if verbPos == None:
                    if hasattr(c, "isProg") or hasattr(c, "isMod"):
                        if hasattr(c, "isProg"): prog = c
                        i += 1
                        continue
                    verbPos = i
                    # find the appropriate clitic table to use
                    t = c.getProp("t")
                    if t == "ip":
                        cliticTable = Constituent.proclitiqueOrdreImperatifNeg if hasattr(c, "neg2") \
                            else Constituent.proclitiqueOrdreImperatifPos
                    elif t == "b":
                        cliticTable = Constituent.proclitiqueOrdreInfinitif
                    else:
                        cliticTable = Constituent.proclitiqueOrdre
                    # check for negation
                    if hasattr(c, "neg2") and c.neg2 is not None:
                        c.insertReal(pros, Adv("ne", "fr"))
                        if t == "b":
                            c.insertReal(pros, Adv(c.neg2, "fr"))
                        else:
                            neg2 = c.neg2
                if c.isReflexive() and c.getProp("t") != "pp":
                    if prog != None: c = prog
                    c.insertReal(pros,
                                 Pro("moi", "fr").c("refl").pe(c.getProp("pe")).n(c.getProp("n")).g(c.getProp("g")))
                i += 1
            elif c.isA("Pro") and verbPos != None:
                if c.getProp("c") in ["refl", "acc", "dat"] or c.lemma == "y" or c.lemma == "en":
                    pros.append(cList.pop(i))
                else:
                    i += 1
                    # HACK: stop when seeing a preposition (except "par" introduced by a passivee) or a conjunction
                    #          or a "strange" pronoun that might start a phrase
                    #       whose structure has been flattened at this stage
            elif c.isOneOf(["P", "C", "Adv", "Pro"]) and verbPos != None and c.lemma != "par":
                break
            else:
                i += 1
        if verbPos == None: return
        # add ending "pas" after the verb unless it is "lié" in which cas it goes after the next word
        if neg2 != None:
            vb = cList[verbPos]
            vb.insertReal(cList, Adv(neg2, "fr"), verbPos + (1 if "lier" not in vb.props else 2))
        if len(pros) > 1:
            pros.sort(key=lambda pro: cliticTable[pro] if pro in cliticTable else 100)
        # insert pronouns before the verb
        cList[verbPos:verbPos] = pros

    # applies to a list of Constituents (can be a single one)
    # adds either to the first or last token (which can be the same)
    def doFormat(self,cList):
        punctuation=getRules()["punctuation"]
        lex=getLexicon()   
    
        def getBeforeAfterString(punct):
            if punct in lex and "Pc" in lex[punct]:
                if "compl" in lex[punct]["Pc"]:
                    compl=lex[punct]["Pc"]["compl"]
                    before=punct
                    after=compl
                    tab=lex[punct]["Pc"]["tab"]
                    tabBefore=tab[0]
                    tabAfter=tab[1] if len(tab)==2 else lex[compl]["Pc"]["tab"][0] # get from table of compl 
                    puncRuleBefore=punctuation[tabBefore]
                    puncRuleAfter=punctuation[tabAfter]
                    return {"b":puncRuleBefore["b"]+before+puncRuleBefore["a"],
                            "a":puncRuleAfter["b"]+after+puncRuleAfter["a"]}
                else:
                    tab=lex[punct]["Pc"]["tab"][0]
                    puncRule=punctuation[tab]
                    punct=puncRule["b"]+punct+puncRule["a"]
            return {"b":punct,"a":punct}
    
        # add before the first element of cList and after the last element of cList
        def wrapWith(before,after):
            cList[0].realization=before+cList[0].realization
            cList[-1].realization+=after

        def startTag(tagName, attrs=None):
            if attrs is None:attrs = {}
            attString="".join(f' {key}="{val}"' for key,val in attrs.items())
            return "<"+tagName+attString+">"
    
        # remove possible empty realization strings (often generated by D("a").n("p")) which can break elision
        # but ensure there is at least one left so that options (.en, .a, .b) can be added.
        def removeEmpty(cList):
            i=0
            while i<len(cList):
                if cList[i].realization=="" and len(cList)>1: del cList[i]
                else:
                    i+=1
    
        # start of processing
        removeEmpty(cList)
        # reorder French pronouns within a VP
        if self.isFr() and (self.isA("VP") or (self.isOneOf(deprels) and self.terminal.isA("V"))):
            self.doFrenchPronounPlacement(cList)
            
        if self.isFr():
            self.doElisionFr(cList)
        else:
            self.doElisionEn(cList)
    
        if "cap" in self.props and self.props["cap"]!=False:
            cList[0].realization=cList[0].realization.capitalize()
 
        if "a" in self.props:
            for a in self.props["a"]:
                wrapWith("",getBeforeAfterString(a)["b"])
 
        if "b" in self.props:
            for b in self.props["b"]:
                wrapWith(getBeforeAfterString(b)["b"],"")

        if "en" in self.props or "ba" in self.props:
            ens= self.props["en"] if "en" in self.props else self.props["ba"]
            for en in ens:
                ba=getBeforeAfterString(en)
                wrapWith(ba["b"],ba["a"])
        
        if "tag" in self.props:
            for attName,attVal in self.props["tag"]:
                wrapWith(startTag(attName,attVal),"</"+attName+">")

        return cList
    
    def detokenize(self,terminals):
        if len(terminals)==0:return ""
        s=""
        for i in range(0,len(terminals)-1):
            terminal=terminals[i]
            if "lier" in terminal.props:
                s+=terminal.realization+"-"
                # check for adding -t- in French between a verb and a pronoun
                if self.isFr() and terminal.isA("V") and terminals[i+1].isA("Pro"):
                    # According to Antidote:
                    # C’est le cas, notamment, quand le verbe à la 3e personne du singulier du passé, du présent ou 
                    # du futur de l’indicatif se termine par une autre lettre que d ou t et qu’il est suivi 
                    # des pronoms sujets il, elle ou on. Dans ce cas, on ajoute un ‑t‑ entre le verbe 
                    # et le pronom sujet inversé.
                    if re.search(r"[^dt]$",terminal.realization) and re.match(r"[ieo]",terminals[i+1].realization):
                        s+="t-"
            elif re.search(r"[- ']$",terminal.realization):
                s+=terminal.realization
            elif len(terminal.realization)>0:
                s+=terminal.realization+" "
        s+=terminals[-1].realization
        # apply capitalization and final full stop unless .cap(False)
        if self.parentConst is None:
            if self.isOneOf(["S","root"]) and len(s)>0: ## this is a top-level S
                if "cap" not in self.props or self.props["cap"]!=False:
                    sepWordRE=sepWordREen if self.isEn() else sepWordREfr
                    m=sepWordRE.match(s)
                    idx=len(m.group(1)) # get index of first letter
                    if idx<len(s): # check if there was a letter
                        s=s[0:idx]+s[idx].upper()+s[idx+1:]
                    if "tag" not in self.props or self.props["tag"]==False: # do not touch top-level tag
                        # and a full stop at the end unless there is already one
                        # taking into account any trailing HTML tag
                        m=re.search(r"(.)( |(<[^>]+>))*$",s)
                        if m!=None and m.group(1) not in "?!.:;/":
                            s+=". "   # add a space after "." , like for rule "pc4"
        return s
    
    ## this looks very simple, but it is the start of the realization process
    def realize(self) -> str:
        terminals=self.real()
        return self.detokenize(terminals)
    
    def __str__(self):
        ## in Javascript, the realization is implicit with .toString(),
        ## so in Python, it should be also implicit with str(), but
        ## the Python debugger in PyCharm also uses str() to display information
        ## which can in some cases change the structure.
        ## This makes it hard to follow when tracing... a classic case of Heisenbug!
        ## I found more useful during development to (un)comment the following two lines
        ## but then realization must be launched with ".realize()"
        # return self.toSource(-1)
        return self.realize()
    
    def clone(self,env=None):
        return eval(self.toSource(),globals() if env is None else env)
    
    def toSource(self):
        return self.optSource
    
    def setJSONprops(self,json):
        if "props" in json:
            for opt,val in json["props"].items():
                if hasattr(self,opt):
                    if isinstance(val,list):
                        for o in val:
                            if isinstance(o,list):
                                getattr(self,opt)(*o)
                            else:
                                getattr(self,opt)(o)
                    elif opt not in ["pat","h"]: # do not copy the pat or h properties of a verb
                        getattr(self,opt)(val)
                else:
                    print("Constituent.fromJSON: illegal prop:"+opt,file=sys.stderr)
        return self                    
    
    def warn(self,*args):
        from .Warning import warning
        # print("WARNING:",self.me(),args,file=sys.stderr)
        print(warning(self,args),file=sys.stderr)
        return self


    
## add Constituent methods for setting properties 
## adapted from https://stackoverflow.com/questions/8276733/dynamically-add-methods-to-a-class-in-python-3-0
# for standard options
def makeOptionMethod(option,validVals,allowedConsts,optionName=None):
    def _method(self, val=None,prog=None):
        nonlocal optionName
        if val is None:
            if validVals!=None and "" not in validVals:
                return self.warn("no value for option",option,validVals)
        if self.isA("CP") and option not in ["cap","lier"]:
            # propagate an option through the children of a CP except for "cap" and "lier"
            if prog is None:self.addOptSource(optionName,val)
            for e in self.elements:
                if len(allowedConsts)==0 or e.isOneOf(allowedConsts):
                    getattr(e,option)(val)
            return self
        if len(allowedConsts)==0 or self.isOneOf(allowedConsts) or self.isOneOf(deprels):
            if validVals!=None and val not in validVals:
                return self.warn("ignored value for option",option,val)
            # start of the real work
            if optionName is None:optionName=option
            self.setProp(optionName,val)
            if prog is None:self.addOptSource(option,val)
            return self
        else:
            if quoteOOV and self.isA("Q"):return self
            return self.warn("bad const for option",option,self.constType,allowedConsts)
        
    return _method


# shared properties 
#   pe,n and g : can be applied to components of NP and Sentences
setattr(Constituent,"pe",makeOptionMethod("pe",[1,2,3,'1','2','3'],["D","Pro","N","NP","A","AP","V","VP","S","SP","CP"]))
setattr(Constituent,"n",makeOptionMethod("n",["s","p","x"],["D","Pro","N","NP","A","AP","V","VP","S","SP","CP"],None))
setattr(Constituent,"g",makeOptionMethod("g",["m","f","n","x"],["D","Pro","N","NP","A","AP","V","VP","S","SP","CP"]))
#  t, aux : can be applied to VP and sentence
setattr(Constituent,"t",makeOptionMethod("t",["p", "i", "f", "ps", "c", "s", "si", "ip", "pr", "pp", "b", # simple tenses
                   "pc", "pq", "cp", "fa", "spa", "spq"],["V","VP","S","SP","CP"]))  # composed tenses
setattr(Constituent,"aux",makeOptionMethod("aux",["av","êt","aê"],["V","VP","S","SP","CP"]))
# ordinary properties
setattr(Constituent,"f",makeOptionMethod("f",["co","su"],["A","Adv"]))
setattr(Constituent,"tn",makeOptionMethod("tn",["","refl"],["Pro"]))
setattr(Constituent,"c",makeOptionMethod("c",["nom","acc","dat","refl","gen"],["Pro"]))

setattr(Constituent,"pos",makeOptionMethod("pos",["post","pre"],["A"]))
setattr(Constituent,"pro",makeOptionMethod("pro",None,["NP","PP"]))
# English only
setattr(Constituent,"ow",makeOptionMethod("ow",["s","p","x"],["D","Pro"],"own"))

# Formatting options
setattr(Constituent,"cap",makeOptionMethod("cap",None,[]))
setattr(Constituent,"lier",makeOptionMethod("lier",None,[]))

def makeOptionListMethod(option):
    def _method(self,val,prog=None):
        if option not in self.props:self.props[option] = []
        self.props[option].append(val)
        if prog is None:self.addOptSource(option,val)
        return self
    return _method

for name in optionListMethods:
    setattr(Constituent, name, makeOptionListMethod(name))

      


from .Constituent import Constituent,quoteOOV
from .Number import enToutesLettres, ordinal
from .Lexicon import getLexicon,currentLanguage,getLemma,getRules

import datetime, re

class Terminal(Constituent):
    def __init__(self, terminalType,lemma,lang=None):
        super().__init__(terminalType)
        self.lemma=lemma # needed for error message
        if lang is None:
            self.lang=currentLanguage()
        elif lang in ["en","fr"]:
            self.lang=lang
        else:
            self.lang=currentLanguage()
            self.warn("bad language",self.lang)
        self.setLemma(lemma,terminalType)
            
    def me(self):
        return self.constType+"("+repr(self.lemma)+")"
    
    def morphoError(self,errorKind,keyVals):
        self.warn("morphology error",errorKind+f" {self.me()} :"+repr(keyVals))
        self.realization=f"[[{self.lemma}]]"
        self.constType="Q"
        return self
    
    ## TODO: see Warning.js
    def add(self,_):
        self.warn("bad application",".add","Phrase",type(self))
        return self
    
    # set lemma, precompute stem and store conjugation/declension table number 
    # TODO: add lang as last parameter
    def setLemma(self,lemma,terminalType=None):
        if terminalType is None: # when it is not called from a Constructor, keep the current terminalType
            terminalType=type(self).__name__
        if isinstance(lemma, str):
            lemma=lemma.replace("œ","oe").replace("æ","ae")
        self.lemma=lemma
        if not hasattr(self, "peng"):self.initProps()
        if terminalType=="DT":
            if lemma is None or lemma=="":
                self.date=datetime.datetime.today()
            else:
                if isinstance(lemma,str):
                    self.date=self.parseDateString(lemma)
                elif isinstance(lemma,datetime.datetime):
                    self.date=lemma
                else:
                    self.warn("bad parameter","str,datetime.datetime",type(lemma).__name__)
                    self.date=datetime.datetime.today()                   
            self.lemma=str(self.date)
            ## set defaults
            self.props["dOpt"]={"year":True,"month":True,"date":True,"day":True,
                        "hour":True,"minute":True,"second":True,
                        "nat":True,"det":True,"rtime":False}
        elif terminalType=="NO":
            if not isinstance(lemma,(str,int,float)):
                self.warn("bad parameter",["str","int","float"],type(lemma).__name__)
                self.lemma=0
            elif isinstance(lemma,str):
                # check if this looks like a legal number
                if not re.match(r"^[-+]?[0-9]+([., ][0-9]*)?([Ee][-+][0-9]+)?$",lemma):
                    self.warn("bad parameter","number",type(lemma).__name__)
                    self.lemma=0
                else:
                    self.lemma=re.sub(r"," if self.isEn() else r" ","",self.lemma)
                try:
                    self.value=int(self.lemma)
                except ValueError:
                    self.value=float(self.lemma)
                self.nbDecimals=self.lemma[::-1].find(".")
            else:
                self.value=lemma
                self.nbDecimals=str(lemma)[::-1].find(".")
            if self.nbDecimals<0:self.nbDecimals=0
            self.props["dOpt"]={"mprecision":2,"raw":False,"ord":False}
        elif terminalType=="Q":
            self.lemma=str(lemma)
        elif terminalType in ["N","A","Pro","D","V","Adv","C","P"]:
            if not isinstance(lemma,str):
                self.tab=None 
                self.realization=f"[[{lemma}]]"
                return self.warn("bad parameter","string",type(lemma).__name__)
            lexInfo=getLemma(self.lemma)
            if lexInfo is None:
                self.tab=None 
                self.realization=f"[[{lemma}]]"
                return self.warn("not in lexicon",self.lang)
                if quoteOOV: # not currently used
                    self.lemma=str(lemma)
                    self.realization=self.lemma
                    self.constType="Q"
            else:
                if terminalType not in lexInfo:
                    self.tab=None 
                    self.realization=f"[[{lemma}]]"
                    return self.warn("not in lexicon",self.lang,lexInfo.keys())
                    if quoteOOV: # not currently used
                        self.lemma=str(lemma)
                        self.realization=self.lemma
                        self.constType="Q"
                else:
                    lexInfo=lexInfo[terminalType]
                    rules=getRules()
                    for (key,info) in lexInfo.items():
                        if key=="tab": # save table number and compute stem
                            ending=None 
                            self.tab=info
                            if not self.isA("V"):
                                if self.tab in rules["declension"]: #  occurs for C, Adv and P
                                    declension=rules["declension"][self.tab]
                                    ending=declension["ending"]
                                    if terminalType=="Pro":
                                        #  set person for Pro when different than 3 (i.e. all elements of
                                        #  declension are the same)
                                        dd=declension["declension"]
                                        if "pe" in dd[0]:
                                            pe=dd[0]["pe"]
                                            if pe!=3:
                                                i=1
                                                while i<len(dd) and dd[i]["pe"]==pe:i+=1
                                                if i==len(dd):self.setProp("pe",pe)
                                        else:
                                            pe=3
                            else:
                                if self.tab in rules["conjugation"]:
                                    ending=rules["conjugation"][self.tab]["ending"]
                                else:
                                    ending=""
                                    self.warn("bad lexicon table",lemma,ending)
                            if ending is not None and self.lemma.endswith(ending):
                                self.stem=self.lemma if ending=="" else self.lemma[:-len(ending)]
                            else:
                                self.tab=None 
                                if not self.isOneOf(["Adv","C","P"]):
                                    self.warn("bad lexicon table",lemma,ending)
                        else:
                            # if isinstance(info,list) and len(info)==1:info=info[0]
                            self.setProp(key,info)
        else:
            self.warn("not implemented",terminalType)
    
    def grammaticalNumber(self):
        if not self.isA("NO"):
            return self.warn("bad application","grammaticalNumber","NO",self.constType)
        if "ord" in self.props and self.props["ord"]==True:
            return "s"
        number=self.value
        if self.isFr():
            # according to http:#bdl.oqlf.gouv.qc.ca/bdl/gabarit_bdl.asp?id=1582
            return "s" if -2<number<2 else "p"
        else:
            # according to https:#www.chicagomanualofstyle.org/book/ed17/part2/ch09/psec019.html
            #   any number other than 1 is plural... 
            # even 1.0 but this case is not handled here because nbDecimal(1.0)=>0
            return "s" if abs(number)==1 and self.nbDecimals==0 else "p"
    
    def getIndex(self,constTypes):
        if isinstance(constTypes, str):
            return 0 if self.isA(constTypes) else -1
        return 0 if self.isOneOf(constTypes) else -1
    
    def getConst(self,constTypes):
        return self if self.getIndex(constTypes)==0 else None 
    
    # try to find the best declension match
    #    value equal = 2
    #    equal with x = 1
    #    no match = 0
    #  but if the person does not match set score to 0           
    def bestMatch(self,errorKind,declension,keyVals):                       
        bestMatch=(0,None)
        for d in declension:
            nbMatches=0
            for key,val in keyVals.items():
                if key in d:
                    if key=="pe" and d[key]!=val: # persons must match exactly
                        nbMatches=0
                        break
                    if d[key]==val:nbMatches+=2
                    elif d[key]=="x":nbMatches+=1
            if nbMatches>bestMatch[0]:
                bestMatch=(nbMatches,d["val"])
        if bestMatch[0]==0:
            self.morphoError(errorKind, keyVals)
            return None 
        return bestMatch[1]                 
                                     
    def decline(self,setPerson):
        rules=getRules()
        declension=rules["declension"][self.tab]["declension"]
        stem=self.stem
        res=None 
        if self.isOneOf(["A","Adv"]):
            if self.isFr():
                g = self.getProp("g")
                n = self.getProp("n")
                ending=self.bestMatch("déclinaison d'adjectif", declension,{"g":g,"n":n})
                if ending is None:
                    return f"[[{self.lemma}]]"
                res=stem+ending
                f = self.getProp("f") # comparatif d'adjectif
                if f is not None and f != False:
                    specialFRcomp={"bon":"meilleur","mauvais":"pire"}
                    if f=="co":
                        if self.lemma in specialFRcomp:
                            return A(specialFRcomp[self.lemma]).g(g).n(n).realize()
                        else:
                            return "plus "+res
                    if f=="su":
                        art=D("le").g(g).n(n).realize()+" "
                        if self.lemma in specialFRcomp:
                            return art+A(specialFRcomp[self.lemma]).g(g).n(n).realize()
                        else:
                            return art+"plus "+res
            else: 
                # English adjective/adverbs are invariable but they can have comparative
                res=self.lemma
                f = self.getProp("f")
                if f is not None and f!=False:
                    if self.tab=="a1":
                        res=("more " if f=="co" else "most ")+res
                    else:
                        if self.tab=="b1": # adverb without comparative/superlative, try the adjective table
                            adjAdv=getLemma(self.lemma)
                            if "A" in adjAdv:
                                declension=rules["declension"][adjAdv["A"]["tab"]]["declension"]
                                ending=rules["declension"][adjAdv["A"]["tab"]]["ending"]
                                stem=stem[0:-len(ending)]
                            else:  # adverb without adjective
                                return res
                        # look in the adjective declension table
                        ending=self.bestMatch("adjective declension",declension,{"f":f})
                        if ending is None:
                            return f"[[{self.lemma}]]"
                        res=stem+ending
        elif len(declension)==1: # no declension
            res=stem+declension[0]["val"]
        else: # for N,D,Pro
            g=self.getProp("g")
            if self.isOneOf(["D","N"]) and g is None:g="m"
            n=self.getProp("n")
            if self.isOneOf(["D","N"]) and n is None:n="s"
            pe=3
            if setPerson:
                p=self.getProp("pe")
                pe=3 if p is None else int(p)
            keyVals={"pe":pe,"g":g,"n":n} if setPerson else {"g":g,"n":n}
            if "own" in self.props:
                keyVals["own"]=self.props["own"]
            if self.isA("Pro"):
                c  = self.props["c"] if "c" in self.props else None
                if c is not None:
                    if self.isFr() and c=="gen": # genitive cannot be used in French
                        self.warn("ignored value for option","c",c)
                    elif self.isEn() and c=="refl": # reflechi cannot be used in English
                        self.warn("ignored value for option","c",c)
                    else:
                        keyVals["c"]=c
                tn = self.props["tn"] if "tn" in self.props else None
                if tn  is not None:
                    if c is not None:
                        self.warn("both tonic and clitic")
                    else:
                        keyVals["tn"]=tn
                if c  is not None or tn  is not None:
                    if (self.isFr() and self.lemma=="moi") or  (self.isEn() and self.lemma=="me"):
                        # HACK:remove defaults from pronoun such as "moi" in French and "me" in English
                        #      because their definition is special in order to try to keep some upward compatibility
                        #      with the original way of specifying the pronouns
                        if self.getProp("g")  is None: del keyVals["g"]
                        if self.getProp("n")  is None: del keyVals["n"]
                        # make sure it matches the first and set the property for verb agreement
                        if (c=="nom" or tn=="") and self.getProp("pe") is None:
                            keyVals["pe"]=1
                            self.setProp("pe",1)
                    else: # set person, gender and number except when subject in an English genitive
                        d0=declension[0]
                        if self.isFr() or c != "gen":
                            self.setProp("g", d0["g"] if "g" in d0 else g)
                            self.setProp("n", d0["n"] if "n" in d0 else n)
                            keyVals["pe"] = d0["pe"] if "pe" in d0 else 3
                            self.setProp("pe",keyVals["pe"])
                else: # no c, nor tn set tn to "" except for "on"
                    if self.lemma!="on": keyVals["tn"]=""
            ending=self.bestMatch("déclinaison" if self.isFr() else "declension",declension,keyVals)
            if ending is None:
                return f"[[{self.lemma}]]"
            if self.isFr() and self.isA("N"):
                # check is French noun gender specified corresponds to the one given in the lexicon
                lexiconG=getLexicon()[self.lemma]["N"]
                if "g" not in lexiconG:
                    self.morphoError("absent du lexique",{"g":g,"n":n})
                    return f"[[{self.lemma}]]"
                lexiconG=lexiconG["g"]
                if lexiconG != "x" and lexiconG != g:
                    self.morphoError(
                        "genre différent de celui du lexique",{"g":g, "lexique":lexiconG})
                    return f"[[{self.lemma}]]"
            res=stem+ending
        return res

    def removeNextConstInSentence(self):
        parentElems=self.parentConst.elements
        try:
            myIndex=parentElems.index(self)
            nextWord=parentElems[myIndex+1]
            del parentElems[myIndex+1]
            nextWord.realization=nextWord.realize()
            return nextWord
        except:
            return self.error("no parent for removeNextConstInSentence")
    
    # fill the realization field of a constituent
    # used heavily in conjugate_fr and conjugate_en       
    # def constReal(self,cnst):
    #     cnst.realization=cnst.lemma if cnst.isA("Q") else cnst.realize()
    #     return cnst
    
    # insert a pyrealb terminal with its realization field already filled in a list of terminal
    # used heavily in conjugate_fr and conjugate_en
    def insertReal(self,terms,newTerminal,position=None):
        if isinstance(newTerminal,Terminal):
            newTerminal.parentConst=self.parentConst
            newTerminal.realize()
            if position is None:
                terms.append(newTerminal)
            else:
                terms.insert(position,newTerminal)
        else:
            self.warn("bad Constituent",NO(position+1).dOpt({"ord":True}),type(newTerminal).__name__)
        return terms
    
    def isReflexive(self):
        if not self.isA("V"):
            return self.error("isReflexive() should be called only for a verb,  not a "+self.constType)
        if hasattr(self,"ignoreRefl"):return False # HACK: this might be set in Phrase.processTyp_fr when
                                                   # dealing with "progressive"
        pat=self.getProp("pat")
        if pat is not None and len(pat)==1 and pat[0]=="réfl": return True # essentiellement réflexif
        # check for "refl" typ (only called for V): Terminal.conjugate_fr
        pc=self.parentConst
        while pc is not None:
            if pc.isOneOf(["VP","SP","S"]):
                if "typ" in pc.props:
                    typs=pc.props["typ"]
                    if "refl" in typs and typs["refl"]==True:
                        if "réfl" not in pat:
                            self.ignoreRefl=True
                            if self.lemma not in ["avoir","être","pouvoir","devoir","vouloir"]:
                                self.warn("ignored reflexive",pat)
                            return False
                        return True
            pc=pc.parentConst
        return False

    
    def conjugate_fr(self):
        pe=int(self.getProp("pe"))
        g = self.getProp("g")
        n = self.getProp("n")
        t = self.getProp("t")
        if self.tab is None:
            return [self.morphoError("conjugate_fr:tab",{"pe":pe,"n":n,"t":t})] 
        if t in ["pc","pq","cp","fa","spa","spq"]: # temps composés
            tempsAux={"pc":"p","pq":"i","cp":"c","fa":"f","spa":"s","spq":"si"}[t]
            aux=V("avoir","fr")
            aux.parentConst=self.parentConst
            aux.peng=self.peng
            aux.taux={"t":self.taux["t"],"aux":self.taux["aux"]} # separate tense of the auxiliary from the original
            if self.isReflexive():
                aux.setLemma("être")  # réflexive verbs must use French "être" auxiliary
                aux.setProp("pat",["réfl"]) # set reflexive for the auxiliary
            elif aux.taux["aux"]=="êt":
                aux.setLemma("être")
            else: # auxiliary "avoir"
                # check the gender and number of a cod appearing before the verb to do proper agreement
                # of its part participle
                g="m"
                n="s"
                if hasattr(self,"cod"):
                    g=self.cod.getProp("g")
                    n=self.cod.getProp("n")
            aux.taux["t"]=tempsAux
            aux.realization=aux.realize()
            # change this verb to pp
            self.setProp("g",g)
            self.setProp("n",n)
            self.setProp("t","pp")
            self.realization=self.realize()
            if hasattr(self,"neg2"):
                aux.neg2=self.neg2 # save this flag to put on the auxiliary, 
                del self.neg2      # delete it on this verb
            if "lier" in self.props:
                aux.lier()         # put this flag on the auxiliary
                del self.props["lier"] # delete it from the verb
                # HACK: check if the verb was lié to a nominative pronoun (e.g. subject inversion for a question)
                myParent=self.parentConst
                if myParent is not None:
                    if hasattr(myParent,"elements"): # when dealing with a Phrase
                        myParentElems=myParent.elements
                        idxMe=myParentElems.index(self)
                        if idxMe>=0 and idxMe<len(myParentElems)-1:
                            idxNext=idxMe+1
                            next=myParentElems[idxNext]
                            if next.isA("Pro"):
                                thePro=myParentElems.pop(idxNext) # remove next pro from parent
                                thePro.realization=thePro.realize() # insert its realization after the auxiliary
                                                                    # and before the verb
                                return [aux,thePro,self]
                    elif hasattr(myParent,"dependents"): # when dealing with a Dependent
                        proIndex=myParent.findIndex(lambda d:d.terminal.isA("Pro"))
                        if proIndex>=0:
                            thePro=myParent.removeDependent(proIndex).terminal  # remove Pro from Parent
                            thePro2=Pro(thePro.lemma)             # as the original Pro is already realized in the
                                                                  # output list, we must hack
                            thePro2.props=thePro.props            # and we cannot use clone because of environ
                            thePro2.peng=thePro.peng
                            thePro2.realization=thePro2.realize() # insert its realization after the auxiliary
                                                                  # and before the verb
                            thePro.realization=""                 # set original Pro realization to nothing
                            return [aux,thePro2,self]
                    else:
                        self.error("Terminal.conjugate_fr:: Strange parent:" + type(myParent).__name__)
            return [aux,self]
        else: # simple tense
            conjugationTable=getRules()["conjugation"][self.tab]
            if "t" in conjugationTable and t in conjugationTable["t"]:
                conjugation=conjugationTable["t"]
                if t in ["p","i","f","ps","c","s","si"]:
                    term=conjugation[t][pe-1+(3 if n=="p" else 0)]
                    if term is None:
                        return [self.morphoError("conjugate_fr",{"pe":pe,"n":n,"t":t})]
                    else:
                        self.realization=self.stem+term
                    res=[self]
                    if self.parentConst is None and self.isReflexive():
                        self.insertReal(res, Pro("moi","fr").c("refl").pe(pe).n(n).g(g),0)
                    return res
                elif t=="ip":
                    if (n=="s" and pe!=2) or (n=="p" and pe==3):
                        return [self.morphoError("conjugate_fr",{"pe":pe,"n":n,"t":t})]
                    term=conjugation[t][pe-1+(3 if n=="p" else 0)]
                    if term is None:
                        return [self.morphoError("conjugate_fr",{"pe":pe,"n":n,"t":t})]
                    else:
                        self.realization=self.stem+term
                    res=[self]
                    if self.isReflexive() and self.parentConst is None:
                        self.lier()
                        self.insertReal(res,Pro("moi","fr").tn("").pe(pe).n(n).g(g))
                    return res
                elif t in ["b","pr","pp"]:
                    self.realization=self.stem+conjugation[t]
                    res=[self]
                    if t!="pp" and self.isReflexive() and self.parentConst is None:
                        self.insertReal(res, Pro("moi","fr").c("refl").pe(pe).n(n).g(g), 0)
                    if t=="pp" and self.realization!="été": # HACK: frequent case of être that does not change
                        g=self.getProp("g")
                        if g=="x":g="m"
                        n=self.getProp("n")
                        if n=="x":n="s"
                        term={"ms":"","mp":"s","fs":"e","fp":"es"}[g+n]
                        self.realization+=term
                    # neg=self.neg2 if hasattr(self,"neg2") else None
                    # if neg is not None and neg!="":
                    #     if t=="b" or t=="pp":
                    #         self.insertReal(res,Adv(neg,"fr"),0)
                    #     else:
                    #         self.insertReal(res,Adv(neg,"fr"))
                    return res
                else:
                    return [self.morphoError("conjugate_fr",{"pe":pe,"n":n,"t":t})]
            else:
                return [self.morphoError("conjugate_fr",{"pe":pe,"n":n,"t":t})]
    
    def conjugate_en(self):
        pe=self.getProp("pe")
        pe = 3 if pe is None else int(pe)
        n = self.getProp("n")
        t = self.getProp("t")
        if self.tab is None:
            return [self.morphoError("conjugate_en:tab",{"pe":pe,"n":n,"t":t})] 
        # subjonctive present is like present except that it does not end in s at 3rd person
        # subjonctive past is like simple past
        t1 = "p" if t=="s" else ("ps" if t=="si" else t)
        conjugationTable=getRules()["conjugation"][self.tab]
        res=[self]
        if "t" in conjugationTable and t1 in conjugationTable["t"]:
            conjugation=conjugationTable["t"][t1]
            if t in ["p","ps","s","si"]:
                if isinstance(conjugation,str):
                    self.realization=self.stem+conjugation
                    return [self]
                term=conjugation[pe-1+(3 if n=="p" else 0)]
                if term is None:
                    return [self.morphoError("conjugate_en:tab",{"pe":pe,"n":n,"t":t})]
                else:
                    # remove final s at subjonctive present by taking the form at the first person
                    if t=="s" and pe==3:term=conjugation[0]
                    self.realization=self.stem+term
                    return [self]
            elif t in ["b","pp","pr"]:
                self.realization=self.stem+conjugation
                return [self]
        elif t=="f":
            self.realization=self.lemma
            self.insertReal(res,Q("will"),0)
        elif t=="ip":
            self.realization=self.lemma
        else:
            return [self.morphoError("conjugate_en:tab",{"pe":pe,"n":n,"t":t})]
        return res 
        
                
    def conjugate(self):
        return self.conjugate_fr() if self.isFr() else self.conjugate_en()       

    ### Number 
    ###   implementation differes from the javascript one using format 
    ###    we avoid using locale...
    def numberFormatter(self,maxPrecision=None):
        if isinstance(self.value,int):
            precision=0
        else:
            precision = 2 if maxPrecision  is None else maxPrecision
        res=("{:,."+str(precision)+"f}").format(self.value)
        symbol = getRules()["number"]["symbol"]
        if symbol["group"]!=",":res=res.replace(",",symbol["group"])
        if symbol["decimal"]!=".":res=res.replace(".",symbol["decimal"]) 
        return res
        
    def numberToWord(self):
        if not isinstance(self.value,int):
            self.warn("bad number in word",self.value)
            return str(self.value)
        if self.isFr() and self.peng["g"]=="f":
            if self.value==1:return "une"
            if self.value==-1:return "moins une"
        return enToutesLettres(self.value,self.lang)

    def numberToOrdinal(self):
        if not isinstance(self.value,int) or self.value<0:
            self.warn("bad ordinal",self.value)
            return str(self.value)
        return ordinal(self.value,self.lang,self.peng["g"])
    
    ### Date
    def dateFormat(self,dateObj,dOpts):
        fmtRE=re.compile(r"(.*?)\[(.+?)]|(.+$)")
        dateRule = getRules()["date"]
        fmts=dateRule["format"]["natural" if dOpts["nat"] else "non_natural"]

        def interpret(fields):
            if len(fields)==0:return ""
            res=""
            fmt=fmts[fields]
            if "det" in dOpts and not dOpts["det"]:
                fmt=fmt[fmt.index("["):] # remove determiner before the first left bracket
            for m in fmtRE.finditer(fmt):
                if m[1] is None:
                    res+=m[3]
                else:
                    res+=m[1]+{
                            "Y" :lambda:str(dateObj.year),
                            "F" :lambda:dateRule["text"]["month"][str(dateObj.month)],
                            "M0":lambda:f"{dateObj.month:02}",
                            "M" :lambda:str(dateObj.month),
                            "d0":lambda:f"{dateObj.day:02}",
                            "d" :lambda:str(dateObj.day),
                            "l" :lambda:dateRule["text"]["weekday"][(dateObj.weekday()+1)%7],
                            "A" :lambda:dateRule["text"]["meridiem"][0 if dateObj.hour<12 else 1],
                            "h" :lambda:str(dateObj.hour%12),
                            "H0":lambda:f"{dateObj.hour:02}",
                            "H" :lambda:str(dateObj.hour),
                            "m0":lambda:f"{dateObj.minute:02}",
                            "m" :lambda:str(dateObj.minute),
                            "s0":lambda:f"{dateObj.second:02}",
                            "s" :lambda:str(dateObj.second),
                        }[m[2]]()
            return res

        if isinstance(dOpts["rtime"],datetime.datetime):
            relativeDate = dateRule["format"]["relative_time"]
            # find the number of days of difference between relDay and the current date
            diffDays=(dateObj-dOpts["rtime"]).days  # use datetime arithmetic to create timedelta
            if str(diffDays) in relativeDate: # within a week before or after
                dateS = relativeDate[str(diffDays)].replace("[l]",
                            dateRule["text"]["weekday"][(dateObj.weekday()+1)%7])
            else:
                ## more than a week
                sign="-" if diffDays<0 else "+"
                dateS = relativeDate[sign].replace("[x]",str(abs(diffDays)))
        else:
            ## process date fields
            dateS = interpret("-".join(field for field in ["year", "month", "date", "day"] if dOpts[field]))

        timeS=interpret(":".join(field for field in ["hour","minute","second"] if dOpts[field]))
        return " ".join(s for s in [dateS,timeS] if len(s)>0)



    def real(self):
        if self.isOneOf(["N","A"]):
            if hasattr(self, "tab") and self.tab  is not None:
                self.realization=self.decline(False)
        elif self.isA("Adv"):
            if hasattr(self, "tab") and self.tab  is not None:
                self.realization=self.decline(False)
            else:
                self.realization=self.lemma
        elif self.isOneOf(["C","P","Q"]):
            if self.realization is None:
                self.realization=self.lemma
        elif self.isOneOf(["D","Pro"]):
            if hasattr(self, "tab") and self.tab is not None:
                self.realization=self.decline(True)
        elif self.isA("V"):
            return self.doFormat(self.conjugate())
        elif self.isA("DT"):
            self.realization=self.dateFormat(self.date,self.props["dOpt"])
        elif self.isA("NO"):
            if "dOpt" in self.props:
                opts=self.props["dOpt"]
                if "nat" in opts and opts["nat"]==True:
                    self.realization=self.numberToWord()
                elif "ord" in opts and opts["ord"]==True:
                    self.realization=self.numberToOrdinal()
                elif "raw" in opts and opts["raw"]==False:
                    self.realization=self.numberFormatter(opts["mprecision"])
                else: # opts["raw"]==True
                    self.realization=str(self.value)
            else:
                self.realization=str(self.value)
        else:
            self.error("Terminal.error:"+self.constType+"not implemented")
        return self.doFormat([self])
    
    def toSource(self,_indent=0):
        return f"{self.constType}(\"{self.lemma}\")"+super().toSource()

    def toJSON(self):
        res={"terminal":self.constType,"lemma":self.lemma}
        if self.parentConst is None or self.lang!=self.parentConst.lang: # only indicate when language changes
            res["lang"]=self.lang
        if len(self.props): # do not output empty props
            res["props"]=self.props
        return res
    
    @classmethod
    def fromJSON(cls,constType,json,lang):
        if "lemma" in json:
            return Terminal(constType,json["lemma"],lang).setJSONprops(json)
        else:
            print("Terminal.fromJSON: no lemma found in "+str(json))
        return cls
        
    def toDependent(self,depName=None):
        from .Dependent import Dependent,det
        # HACK: self.parentConst in changed during transformation to refer to the parent dependennt...
        isTopLevel = self.parentConst is None  # we save it to use it at the end
        if self.isOneOf(["D", "NO"]):
            deprel=det(self)
        else:
            deprel= Dependent([self], "root" if depName is None else depName)
        if isTopLevel: deprel.cap(False)
        return deprel


# # create Terminal subclasses
#  which could be done using the following exec
# terminalDefString ='''
# class {0}(Terminal):
#     def __init__(self, lemma,lang=None):
#         super().__init__("{0}",lemma)
# '''
# for t in ["N","A","Pro","D","Adv","V","P","C","DT","NO","Q"]:
#     exec(terminalDefString.format(t))
#
# but we use the following in-line expansion of the above to avoid annoying error messages in Eclipse
class N(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("N",lemma,lang)

class A(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("A",lemma,lang)

class Pro(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("Pro",lemma,lang)

class D(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("D",lemma,lang)

class Adv(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("Adv",lemma,lang)

class V(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("V",lemma,lang)

class P(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("P",lemma,lang)

class C(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("C",lemma,lang)

class DT(Terminal):
    def __init__(self, lemma=None,_lang=None):
        super().__init__("DT", lemma) 
            
class NO(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("NO",lemma,lang)

class Q(Terminal):
    def __init__(self, lemma=None,lang=None):
        super().__init__("Q",lemma,lang)
